# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from pathlib import Path
import re

###############################################################################
# Constants
###############################################################################

HERE = Path(__file__).resolve(strict=True).parent
PKG = HERE.parent / 'src' / 'hpl'
GRAMMARS = PKG / 'grammars'

PREAMBLE = re.compile(r'\s*//\s*SPDX-License-Identifier:[^\n]+\s*//\s*Copyright[^\n]+\s*')


def skip_preamble(text: str) -> str:
    m = PREAMBLE.match(text)
    if m:
        i = m.end()
        return text[i:]
    return text


path = GRAMMARS / 'tokens.lark'
g_tokens = skip_preamble(path.read_text(encoding='utf8'))

path = GRAMMARS / 'predicates.lark'
g_predicates = skip_preamble(path.read_text(encoding='utf8'))

path = GRAMMARS / 'properties.lark'
g_properties = skip_preamble(path.read_text(encoding='utf8'))

path = GRAMMARS / 'files.lark'
g_files = skip_preamble(path.read_text(encoding='utf8'))


def get_keyword_list(token_grammar: str) -> str:
    token_map = {}
    pattern = r'(\w+_OPERATOR)(?:\.\d+)?\s*:\s*"(\w+?)"'
    for groups in re.findall(pattern, token_grammar):
        var, token = groups
        token_map[var] = token
    pattern = r'(\w+_OPERATOR)(?:\.\d+)?\s*:\s*/(?:\\b)?(\w+?)(?:\\b)?/'
    for groups in re.findall(pattern, token_grammar):
        var, token = groups
        token_map[var] = token
    return '\n'.join(f"{key} = '{value}'" for key, value in token_map.items())


# def distorted_tokens(token_grammar: str) -> str:
#     def repl(m: re.Match) -> str:
#         return f'{m.group(1)}: " {m.group(2)} "'
# 
#     pattern = r'^(\w+(?:\.\d+)?)\s*:\s*"(\w+?)"$'
#     g = re.sub(pattern, repl, token_grammar, flags=re.M)
#     pattern = r'^(\w+(?:\.\d+)?)\s*:\s*/\\b(\w+?)\\b/$'
#     return re.sub(pattern, repl, g, flags=re.M)


def distorted_tokens(token_grammar: str) -> str:
    def repl(m: re.Match) -> str:
        return f'" {m.group(1)} "'

    pattern = r'(?:"|/\\b)(\w+?)(?:"|\\b/)'
    return re.sub(pattern, repl, token_grammar)


GRAMMAR_PY = f'''\
# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

PREDICATE_GRAMMAR = r"""
{g_predicates}
{g_tokens}
"""

HPL_GRAMMAR = r"""
{g_files}
{g_properties}
{g_predicates}
{g_tokens}
"""

{get_keyword_list(g_tokens)}
'''

path = PKG / 'grammar.py'
path.write_text(GRAMMAR_PY, encoding='utf8')


# TEST_GRAMMAR_PY = f'''\
# # SPDX-License-Identifier: MIT
# # Copyright © 2023 André Santos
# 
# HPL_GRAMMAR = r"""
# {g_files}
# {g_properties}
# {g_predicates}
# {distorted_tokens(g_tokens)}
# """
# '''

# path = HERE.parent / 'tests' / 'grammar.py'
# path.write_text(TEST_GRAMMAR_PY, encoding='utf8')



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from typing import List, Optional, Tuple, TypeVar, Union

import math

from typeguard import typechecked

from hpl.ast.expressions import (
    And,
    BinaryOperatorDefinition,
    BuiltinBinaryOperator,
    Forall,
    FunctionDefinition,
    HplBinaryOperator,
    HplExpression,
    HplFunctionCall,
    HplLiteral,
    HplQuantifier,
    HplRange,
    HplSet,
    HplThisMessage,
    HplUnaryOperator,
    HplVarReference,
    Implies,
    Not,
    Or,
    is_self_reference,
    is_var_reference,
)
from hpl.ast.predicates import (
    HplContradiction,
    HplPredicate,
    HplPredicateExpression,
    HplVacuousTruth,
    predicate_from_expression,
)
from hpl.ast.properties import HplProperty, HplScope

###############################################################################
# Constants
###############################################################################

P = TypeVar('P', HplPredicate, HplExpression)

INVERSE_OPERATORS = {
    BuiltinBinaryOperator.ADD.value: BuiltinBinaryOperator.ADD.value,
    BuiltinBinaryOperator.MULT.value: BuiltinBinaryOperator.MULT.value,
    BuiltinBinaryOperator.AND.value: BuiltinBinaryOperator.AND.value,
    BuiltinBinaryOperator.OR.value: BuiltinBinaryOperator.OR.value,
    BuiltinBinaryOperator.IFF.value: BuiltinBinaryOperator.IFF.value,
    BuiltinBinaryOperator.ADD.value: BuiltinBinaryOperator.ADD.value,
    BuiltinBinaryOperator.EQ.value: BuiltinBinaryOperator.EQ.value,
    BuiltinBinaryOperator.NEQ.value: BuiltinBinaryOperator.NEQ.value,
    BuiltinBinaryOperator.LT.value: BuiltinBinaryOperator.GT.value,
    BuiltinBinaryOperator.GT.value: BuiltinBinaryOperator.LT.value,
    BuiltinBinaryOperator.LTE.value: BuiltinBinaryOperator.GTE.value,
    BuiltinBinaryOperator.GTE.value: BuiltinBinaryOperator.LTE.value,
}

###############################################################################
# Formula Rewriting
###############################################################################


@typechecked
def replace_this_with_var(predicate_or_expression: P, alias: str) -> P:
    var = HplVarReference(f'@{alias}')
    if predicate_or_expression.is_expression:
        if is_self_reference(predicate_or_expression):
            return var
    return predicate_or_expression.replace_self_reference(var)


@typechecked
def replace_var_with_this(predicate_or_expression: P, alias: str) -> P:
    this = HplThisMessage()
    if predicate_or_expression.is_expression:
        if is_var_reference(predicate_or_expression, alias=alias):
            return this
    return predicate_or_expression.replace_var_reference(alias, this)


@typechecked
def refactor_reference(predicate_or_expression: P, alias: str) -> Tuple[P, P]:
    if predicate_or_expression.is_predicate:
        return _refactor_ref_pred(predicate_or_expression, alias)
    else:
        return _refactor_ref_expr(predicate_or_expression, alias)


@typechecked
def canonical_form(property: HplProperty) -> List[HplProperty]:
    if property.pattern.is_safety:
        return _canonical_form_safety(property)
    assert property.pattern.is_liveness
    return _canonical_form_liveness(property)


@typechecked
def split_and(predicate_or_expression: P) -> List[HplExpression]:
    if predicate_or_expression.is_predicate:
        assert isinstance(predicate_or_expression, HplPredicate)
        return _split_and_expr(predicate_or_expression.condition)
    assert isinstance(predicate_or_expression, HplExpression)
    return _split_and_expr(predicate_or_expression)


@typechecked
def simplify(predicate_or_expression: P) -> P:
    if predicate_or_expression.is_predicate:
        assert isinstance(predicate_or_expression, HplPredicate)
        if predicate_or_expression.is_vacuous:
            return predicate_or_expression
        expr: HplExpression = _simplify(predicate_or_expression.condition)
        if is_true(expr):
            return HplVacuousTruth()
        if is_false(expr):
            return HplContradiction()
        return HplPredicateExpression(expr)
    assert isinstance(predicate_or_expression, HplExpression)
    return _simplify(predicate_or_expression)


###############################################################################
# Formula Rewriting - Helper Functions
###############################################################################


@typechecked
def _refactor_ref_pred(phi: HplPredicate, alias: str) -> Tuple[HplPredicate, HplPredicate]:
    if phi.is_vacuous:
        return (phi, HplVacuousTruth())
    expr1, expr2 = _refactor_ref_expr(phi.condition, alias)
    return (predicate_from_expression(expr1), predicate_from_expression(expr2))


@typechecked
def _refactor_ref_expr(expr: HplExpression, alias: str) -> Tuple[HplExpression, HplExpression]:
    if not expr.contains_reference(alias):
        return (expr, true())
    if not expr.can_be_bool:
        # move the whole expression
        return (true(), expr)
    if expr.is_value or expr.is_accessor or expr.is_function_call:
        # cannot split into two parts
        return (true(), expr)
    if expr.is_quantifier:
        return _split_ref_quantifier(expr, alias)
    if expr.is_operator:
        return _split_ref_operator(expr, alias)
    raise TypeError(f'unknown expression type: {expr!r}')


@typechecked
def _split_ref_quantifier(quant: HplQuantifier, alias: str) -> Tuple[HplExpression, HplExpression]:
    var = quant.variable
    if quant.domain.contains_reference(alias):
        # move the whole expression
        return (true(), quant)
    expr = quant.condition
    assert expr.contains_reference(alias)
    # TODO optimize for nested quantifiers
    if quant.is_universal:
        # (A x: p & q)  ==  ((A x: p) & (A x: q))
        if is_not(expr) and is_or(expr.operand):
            expr = And(Not(expr.operand.a), Not(expr.operand.b))
        if is_and(expr):
            a = expr.a.contains_reference(alias)
            b = expr.b.contains_reference(alias)
            va = expr.a.contains_reference(var)
            vb = expr.b.contains_reference(var)
            if a and not b:
                if va:
                    qa = Forall(var, quant.domain, expr.a)
                else:
                    qa = Or(empty_test(quant.domain), expr.a)
                if vb:
                    qb = Forall(var, quant.domain, expr.b)
                else:
                    qb = Or(empty_test(quant.domain), expr.b)
                return (qb, qa)
            if b and not a:
                if va:
                    qa = Forall(var, quant.domain, expr.a)
                else:
                    qa = Or(empty_test(quant.domain), expr.a)
                if vb:
                    qb = Forall(var, quant.domain, expr.b)
                else:
                    qb = Or(empty_test(quant.domain), expr.b)
                return (qa, qb)
            assert a and b
        # move everything
        return (true(), quant)
    if quant.is_existential:
        # (E x: p -> q)  ==  (E x: ~p | q)
        # (E x: p | q)  ==  ((E x: p) | (E x: q))
        # move everything, not worth splitting disjunctions
        return (true(), quant)
    raise TypeError(f'unknown quantifier type: {quant!r}')


@typechecked
def _split_ref_operator(
    op: Union[HplUnaryOperator, HplBinaryOperator], alias: str
) -> Tuple[HplExpression, HplExpression]:
    if op.arity == 1:
        assert isinstance(op, HplUnaryOperator)
        assert op.operator.is_not
        return _split_ref_negation(op, alias)
    else:
        assert op.arity == 2
        assert isinstance(op, HplBinaryOperator)
        if is_and(op):
            a = op.a.contains_reference(alias)
            b = op.b.contains_reference(alias)
            if a and not b:
                return (op.b, op.a)
            if b and not a:
                return (op.a, op.b)
            assert a and b
        # cannot split into two parts
        return (true(), op)


@typechecked
def _split_ref_negation(neg: HplUnaryOperator, alias: str) -> Tuple[HplExpression, HplExpression]:
    expr = neg.operand
    assert expr.can_be_bool and expr.contains_reference(alias)
    if expr.is_value or expr.is_accessor or expr.is_function_call:
        # cannot split into two parts
        return (true(), neg)
    if expr.is_quantifier:
        assert isinstance(expr, HplQuantifier)
        if expr.is_existential:
            # (~E x: p)  ==  (A x: ~p)
            p = Not(expr.condition)
            assert p.contains_reference(expr.variable)
            expr = Forall(expr.variable, expr.domain, p)
            return _split_ref_quantifier(expr, alias)
        # TODO optimize for other (harder) cases
        return (true(), neg)
    if expr.is_operator:
        if is_not(expr):
            # ~~p  ==  p
            return _refactor_ref_expr(expr.operand, alias)
        if is_implies(expr):
            # ~(a -> b)  ==  ~(~a | b)  ==  a & ~b
            expr = And(expr.a, Not(expr.b))
            return _split_ref_operator(expr, alias)
        if is_or(expr):
            # ~(a | b)  ==  ~a & ~b
            expr = And(Not(expr.a), Not(expr.b))
            return _split_ref_operator(expr, alias)
        # cannot split into two parts
        return (true(), neg)
    raise TypeError(f'unknown expression type: {expr!r}')


@typechecked
def _canonical_form_safety(property: HplProperty) -> List[HplProperty]:
    scopes = _canonical_form_scopes(property.scope)
    pattern = property.pattern
    patterns = [pattern.but(behaviour=event) for event in pattern.behaviour.simple_events()]
    if len(scopes) == 1 and len(patterns) == 1:
        # nothing changed
        return [property]
    return [property.but(scope=scope, pattern=pattern) for scope in scopes for pattern in patterns]


@typechecked
def _canonical_form_liveness(property: HplProperty) -> List[HplProperty]:
    scopes = _canonical_form_scopes(property.scope)
    if property.pattern.is_existence:
        patterns = [property.pattern]  # no splits
    else:
        assert property.pattern.is_response, f'pattern: {property.pattern}'
        trigger = property.pattern.trigger
        assert trigger is not None
        patterns = [property.pattern.but(trigger=event) for event in trigger.simple_events()]
    if len(scopes) == 1 and len(patterns) == 1:
        # nothing changed
        return [property]
    return [property.but(scope=scope, pattern=pattern) for scope in scopes for pattern in patterns]


@typechecked
def _canonical_form_scopes(scope: HplScope) -> List[HplScope]:
    if scope.is_after:  # after or after-until
        assert scope.activator is not None
        return [scope.but(activator=event) for event in scope.activator.simple_events()]
    assert scope.is_global or scope.is_until
    return [scope]


@typechecked
def _split_and_expr(phi: HplExpression) -> List[HplExpression]:
    conditions: List[HplExpression] = []
    stack: List[HplExpression] = [phi]
    while stack:
        expr: HplExpression = stack.pop()
        # preprocessing
        if is_true(expr):
            continue
        if is_false(expr):
            raise ValueError(f'unsatisfiable: {phi}')
        expr = _and_presplit_transform(expr)
        # expr should be either an And or something undivisible
        # splits
        if is_and(expr):
            assert isinstance(expr, HplBinaryOperator)
            stack.append(expr.a)
            stack.append(expr.b)
        else:
            conditions.append(expr)
    return conditions


@typechecked
def _and_presplit_transform(phi: HplExpression) -> HplExpression:
    # This should not need a loop any longer
    # previous = None
    # while phi is not previous:
    #     previous = phi
    #     if is_not(phi):
    #         phi = _split_and_not(phi)
    #     elif phi.is_quantifier:
    #         phi = _split_and_quantifier(phi)

    if is_not(phi):
        return _split_and_not(phi)
    if phi.is_quantifier:
        return _split_and_quantifier(phi)
    return phi  # atomic


@typechecked
def _split_and_not(neg: HplUnaryOperator) -> HplExpression:
    """
    Transform a Negation into either an And or something undivisible
    """
    phi: HplExpression = neg.operand
    if is_not(phi):
        # ~~p  ==  p
        assert isinstance(phi, HplUnaryOperator)
        return _and_presplit_transform(phi.operand)
    if is_or(phi):
        # ~(a | b)  ==  ~a & ~b
        assert isinstance(phi, HplBinaryOperator)
        return And(Not(phi.a), Not(phi.b))
    if is_implies(phi):
        # ~(a -> b)  ==  ~(~a | b)  ==  a & ~b
        return And(phi.a, Not(phi.b))
    if phi.is_quantifier:
        assert isinstance(phi, HplQuantifier)
        if phi.is_existential:
            # (~E x: p)  ==  (A x: ~p)
            p = Not(phi.condition)
            assert p.contains_reference(phi.variable)
            phi = Forall(phi.variable, phi.domain, p)
            return _split_and_quantifier(phi)
    return neg


@typechecked
def _split_and_quantifier(quant: HplQuantifier) -> HplExpression:
    """
    Transform a Quantifier into either an And or something undivisible
    """
    var: str = quant.variable
    phi: HplExpression = quant.condition
    if quant.is_universal:
        # (A x: p & q)  ==  ((A x: p) & (A x: q))
        phi = _and_presplit_transform(phi)
        if is_and(phi):
            assert isinstance(phi, HplBinaryOperator)
            if phi.a.contains_reference(var):
                qa = Forall(var, quant.domain, phi.a)
            else:
                qa = Or(empty_test(quant.domain), phi.a)
            if phi.b.contains_reference(var):
                qb = Forall(var, quant.domain, phi.b)
            else:
                qb = Or(empty_test(quant.domain), phi.b)
            return And(qa, qb)
    elif quant.is_existential:
        # (E x: p -> q)  ==  (E x: ~p | q)
        # (E x: p | q)  ==  ((E x: p) | (E x: q))
        pass  # not worth splitting disjunctions
    return quant  # nothing to do


@typechecked
def _simplify(expr: HplExpression) -> HplExpression:
    if isinstance(expr, HplUnaryOperator):
        return _simplify_unary_operator(expr)
    if isinstance(expr, HplBinaryOperator):
        return _simplify_binary_operator(expr)
    if expr.is_function_call:
        return _simplify_function_call(expr)

    if isinstance(expr, HplSet):
        values = [_simplify(v) for v in expr.values]
        n = len(expr.values)
        value_set = set(values)
        if len(value_set) != n:
            return HplSet(value_set)
        for i in range(n):
            if values[i] is not expr.values[i]:
                return HplSet(values)
        return expr  # all the same

    if isinstance(expr, HplRange):
        lb: HplExpression = _simplify(expr.min_value)
        ub: HplExpression = _simplify(expr.max_value)
        if lb is expr.min_value and ub is expr.max_value:
            return expr  # all the same
        return expr.but(min_value=lb, max_value=ub)

    # TODO inclusion, quantifiers, ...
    return expr


@typechecked
def _simplify_unary_operator(expr: HplUnaryOperator) -> HplExpression:
    if expr.operator.is_not:
        return _simplify_negation(expr)
    if expr.operator.is_minus:
        return _simplify_negative_number(expr)
    return expr


@typechecked
def _simplify_negation(phi: HplUnaryOperator) -> HplExpression:
    p: HplExpression = _simplify(phi.operand)
    # ~T == F
    if is_true(p):
        return false()
    # ~F == T
    if is_false(p):
        return true()
    # ~~p  ==  p
    if is_not(p):
        assert isinstance(p, HplUnaryOperator)
        # simplification is redundant, should be recursive
        # return _simplify(p.operand)
        return p.operand
    return phi if p is phi.operand else Not(p)


@typechecked
def _simplify_negative_number(expr: HplUnaryOperator) -> HplExpression:
    a: HplExpression = _simplify(expr.operand)
    if is_number_literal(a):
        assert isinstance(a, HplLiteral)
        return HplLiteral.number(-a.value)
    # --a = a
    if is_negative_number(a):
        assert isinstance(a, HplUnaryOperator)
        # simplification is redundant, should be recursive
        # return _simplify(a.operand)
        return a.operand
    return expr if a is expr.operand else HplUnaryOperator.minus(a)


@typechecked
def _simplify_binary_operator(expr: HplBinaryOperator) -> HplExpression:
    expr = _pre_simplify_binop(expr)
    if expr.operator.is_and:
        return _simplify_conjunction(expr)
    if expr.operator.is_or:
        return _simplify_disjunction(expr)
    if expr.operator.is_implies:
        return _simplify_implies(expr)
    if expr.operator.is_iff:
        return _simplify_iff(expr)
    if expr.operator.is_comparison:
        return _simplify_comparison(expr)
    if expr.operator.is_arithmetic:
        return _simplify_arithmetic(expr)
    if expr.operator.is_inclusion:
        return expr  # FIXME
    return expr


def _pre_simplify_binop(expr: HplBinaryOperator) -> HplBinaryOperator:
    op: BinaryOperatorDefinition = expr.operator
    a: HplExpression = _simplify(expr.operand1)
    b: HplExpression = _simplify(expr.operand2)
    noop: bool = a is expr.operand1 and b is expr.operand2

    # always push literals to the RHS
    if isinstance(b, HplLiteral):
        return expr if noop else HplBinaryOperator(op, a, b)

    # always push self references to the LHS
    if is_self_or_field(a, deep=True):
        return expr if noop else HplBinaryOperator(op, a, b)

    # (1 < x) == (x > 1)
    flip: bool = isinstance(a, HplLiteral)
    # (@y < x) == (x > @y)
    flip = flip or is_self_or_field(b, deep=True)
    if flip:
        if op.commutative:
            return expr.but(operand1=b, operand2=a)
        inv: Optional[BinaryOperatorDefinition] = INVERSE_OPERATORS.get(op)
        if inv is None:
            return expr if noop else HplBinaryOperator(op, a, b)
        return HplBinaryOperator(inv, b, a)

    if op.associative:
        left: bool = isinstance(a, HplBinaryOperator) and a.operator == op
        right: bool = isinstance(b, HplBinaryOperator) and b.operator == op
        if left and right:
            a1: HplExpression = a.operand1
            a2: HplExpression = a.operand2
            b1: HplExpression = b.operand1
            b2: HplExpression = b.operand2
            if isinstance(a2, HplLiteral):
                noop = False
                x = b1
                b1 = b2
                b2 = a2
                a2 = x
                # it is now possible to have redundancy on the LHS
                a = _simplify_binary_operator(HplBinaryOperator(op, a1, a2))
                # it is now possible to have two literals on the RHS
                b = _simplify_binary_operator(HplBinaryOperator(op, b1, b2))
            if is_self_or_field(b1, deep=True):
                noop = False
                x = b1
                b1 = a2
                a2 = a1
                a1 = x
                # it is now possible to have redundancy on the LHS
                a = _simplify_binary_operator(HplBinaryOperator(op, a1, a2))
                # it is now possible to have two literals on the RHS
                b = _simplify_binary_operator(HplBinaryOperator(op, b1, b2))
        elif left:
            if isinstance(a.operand2, HplLiteral):
                noop = False
                x = b
                b = a.operand2
                # it is now possible to have redundancy on the LHS
                a = _simplify_binary_operator(HplBinaryOperator(op, a.operand1, x))
        elif right:
            if is_self_or_field(b.operand1, deep=True):
                noop = False
                x = a
                a = b.operand1
                # it is now possible to have redundancy on the RHS
                b = _simplify_binary_operator(HplBinaryOperator(op, x, b.operand2))

    return expr if noop else HplBinaryOperator(op, a, b)


@typechecked
def _simplify_iff(phi: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    p: HplExpression = phi.operand1
    q: HplExpression = phi.operand2
    # (p == p) == T
    if p == q:
        return true()
    # (p == ~p) == F
    if _obviously_different(p, q):
        return false()
    # (p == q) == ((p -> q) & (q -> p))
    return _simplify(And(Implies(p, q), Implies(q, p)))


@typechecked
def _simplify_implies(phi: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    p: HplExpression = phi.operand1
    q: HplExpression = phi.operand2
    # p -> p == T
    if p == q:
        return true()
    # p -> q == ~p | q
    return _simplify(Or(Not(p), q))


@typechecked
def _simplify_conjunction(phi: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    p: HplExpression = phi.operand1
    q: HplExpression = phi.operand2
    # F & q == F
    if is_false(p):
        return p
    # p & F == F
    if is_false(q):
        return q
    # T & q == q
    if is_true(p):
        return q
    # p & T == p
    if is_true(q):
        return p
    # p & p == p
    if p == q:
        return p
    # p & ~p == F
    if _obviously_different(p, q):
        return false()
    return phi if p is phi.operand1 and q is phi.operand2 else And(p, q)


@typechecked
def _simplify_disjunction(phi: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    p: HplExpression = phi.operand1
    q: HplExpression = phi.operand2
    # T | q == T
    if is_true(p):
        return p
    # p | T == T
    if is_true(q):
        return q
    # F | q == q
    if is_false(p):
        return q
    # p | F == p
    if is_false(q):
        return p
    # p | p == p
    if p == q:
        return p
    # p | ~p == T
    if _obviously_different(p, q):
        return true()
    return phi if p is phi.operand1 and q is phi.operand2 else Or(p, q)


@typechecked
def _simplify_comparison(phi: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    op: BinaryOperatorDefinition = phi.operator
    a: HplExpression = phi.operand1
    b: HplExpression = phi.operand2
    if isinstance(a, HplLiteral) and isinstance(b, HplLiteral):
        if op.is_equality:
            return HplLiteral.boolean(a.value == b.value)
        if op.is_less_than:
            return HplLiteral.boolean(a.value < b.value)
        if op.is_less_than_eq:
            return HplLiteral.boolean(a.value <= b.value)
        if op.is_greater_than:
            return HplLiteral.boolean(a.value > b.value)
        if op.is_greater_than_eq:
            return HplLiteral.boolean(a.value >= b.value)
        assert op.is_inequality
        return HplLiteral.boolean(a.value != b.value)
    if _obviously_different(a, b):
        if op.is_equality:
            return false()
        if op.is_inequality:
            return true()
    return phi


@typechecked
def _simplify_arithmetic(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    op: BinaryOperatorDefinition = expr.operator
    if op.is_plus:
        return _simplify_addition(expr)
    if op.is_minus:
        return _simplify_subtraction(expr)
    if op.is_times:
        return _simplify_multiplication(expr)
    if op.is_division:
        return _simplify_division(expr)
    if op.is_power:
        return _simplify_exponentiation(expr)
    return expr


@typechecked
def _simplify_addition(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    a: HplExpression = expr.operand1
    b: HplExpression = expr.operand2
    if isinstance(b, HplLiteral):
        if b.value == 0:
            return a
        if isinstance(a, HplLiteral):
            return b if a.value == 0 else HplLiteral.number(a.value + b.value)
    if _obvious_negatives(a, b):
        return HplLiteral.number(0)
    return expr


@typechecked
def _simplify_subtraction(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    a: HplExpression = expr.operand1
    b: HplExpression = expr.operand2
    if isinstance(b, HplLiteral):
        if b.value == 0:
            return a
        if isinstance(a, HplLiteral):
            return HplLiteral.number(a.value - b.value)
    if a == b:
        return HplLiteral.number(0)
    if isinstance(b, HplUnaryOperator) and b.operator.is_minus:
        return _simplify_addition(HplBinaryOperator.addition(a, b.operand))
    return expr


@typechecked
def _simplify_multiplication(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    a: HplExpression = expr.operand1
    b: HplExpression = expr.operand2
    if isinstance(b, HplLiteral):
        if b.value == 1:
            return a
        if b.value == 0:
            return b
        if isinstance(a, HplLiteral):
            if a.value == 1:
                return b
            if a.value == 0:
                return a
            return HplLiteral.number(a.value * b.value)
        if b.value == -1:
            return _simplify_unary_operator(HplUnaryOperator.minus(a))
    if isinstance(a, HplBinaryOperator) and a.operator.is_division:
        if a.operand2 == b:
            return a.operand1
    if isinstance(b, HplBinaryOperator) and b.operator.is_division:
        if b.operand2 == a:
            return b.operand1
    # FIXME both are divisions
    return expr


@typechecked
def _simplify_division(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    a: HplExpression = expr.operand1
    b: HplExpression = expr.operand2
    if isinstance(b, HplLiteral):
        if b.value == 0:
            raise ZeroDivisionError(repr(expr))
        if b.value == 1:
            return a
        if isinstance(a, HplLiteral):
            if a.value == 0:
                return a
            return HplLiteral.number(a.value / b.value)
    if a == b:
        return HplLiteral.number(1)
    if _obvious_negatives(a, b):
        return HplLiteral.number(-1)
    return expr


@typechecked
def _simplify_exponentiation(expr: HplBinaryOperator) -> HplExpression:
    # assume simplification already called on operands
    a: HplExpression = expr.operand1
    b: HplExpression = expr.operand2
    if isinstance(b, HplLiteral):
        if b.value == 1:
            return a
        if b.value == 0:
            return HplLiteral.number(1)
        if isinstance(a, HplLiteral):
            if a.value == 1 or a.value == 0:
                return a
            return HplLiteral.number(a.value ** b.value)
    return expr


@typechecked
def _simplify_function_call(call: HplFunctionCall) -> HplExpression:
    fun: FunctionDefinition = call.function
    if fun.name == 'abs':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(abs(arg.value))

    elif fun.name == 'bool':
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplLiteral):
            return HplLiteral.boolean(bool(arg.value))

    elif fun.name == 'int':
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplLiteral):
            return HplLiteral.number(int(arg.value))

    elif fun.name == 'float':
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplLiteral):
            return HplLiteral.number(float(arg.value))

    elif fun.name == 'str':
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplLiteral):
            return HplLiteral.string(str(arg.value))

    elif fun.name == 'len':
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplSet):
            return HplLiteral.number(len(arg.values))
        elif isinstance(arg, HplRange):
            if is_number_literal(arg.min_value) and is_number_literal(arg.max_value):
                n = abs(int(arg.max_value.value) - int(arg.min_value.value))
                if not arg.exclude_max:
                    n += 1
                if arg.exclude_min:
                    n -= 1
            return HplLiteral.number(n)
        elif isinstance(arg, HplLiteral) and isinstance(arg.value, str):
            return HplLiteral.number(len(arg.value))

    elif fun.name == 'sum':
        return _simplify_function_sum(call)

    elif fun.name == 'prod':
        return _simplify_function_prod(call)

    elif fun.name == 'sqrt':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.sqrt(arg.value))

    elif fun.name == 'ceil':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.ceil(arg.value))

    elif fun.name == 'floor':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.floor(arg.value))

    elif fun.name == 'sin':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.sin(arg.value))

    elif fun.name == 'cos':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.cos(arg.value))

    elif fun.name == 'tan':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.tan(arg.value))

    elif fun.name == 'asin':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.asin(arg.value))

    elif fun.name == 'acos':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.acos(arg.value))

    elif fun.name == 'atan':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.atan(arg.value))

    elif fun.name == 'atan2':
        arg1: HplExpression = _simplify(call.arguments[0])
        arg2: HplExpression = _simplify(call.arguments[1])
        if is_number_literal(arg1) and is_number_literal(arg2):
            assert isinstance(arg1, HplLiteral)
            assert isinstance(arg2, HplLiteral)
            return HplLiteral.number(math.atan2(arg1.value, arg2.value))

    elif fun.name == 'deg':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.degrees(arg.value))

    elif fun.name == 'rad':
        arg: HplExpression = _simplify(call.arguments[0])
        if is_number_literal(arg):
            assert isinstance(arg, HplLiteral)
            return HplLiteral.number(math.radians(arg.value))

    elif fun.name == 'log':
        arg1: HplExpression = _simplify(call.arguments[0])
        arg2: HplExpression = _simplify(call.arguments[1])
        if is_number_literal(arg1) and is_number_literal(arg2):
            assert isinstance(arg1, HplLiteral)
            assert isinstance(arg2, HplLiteral)
            if arg2.value == 10:
                return HplLiteral.number(math.log10(arg1.value))
            return HplLiteral.number(math.log(arg1.value, arg2.value))

    elif fun.name == 'max':
        return _simplify_function_max(call)

    elif fun.name == 'min':
        return _simplify_function_min(call)

    elif fun.name == 'gcd':
        # FIXME compound single argument signature
        arg1: HplExpression = _simplify(call.arguments[0])
        arg2: HplExpression = _simplify(call.arguments[1])
        if is_number_literal(arg1) and is_number_literal(arg2):
            assert isinstance(arg1, HplLiteral)
            assert isinstance(arg2, HplLiteral)
            return HplLiteral.number(math.gcd(arg1.value, arg2.value))

    return call


def _simplify_function_sum(call: HplFunctionCall) -> HplExpression:
    arg: HplExpression = _simplify(call.arguments[0])
    if isinstance(arg, HplSet):
        variables: List[HplExpression] = []
        literals: List[Union[int, float]] = []
        for v in arg.values:
            if is_number_literal(v):
                assert isinstance(v, HplLiteral)
                literals.append(v.value)
            else:
                variables.append(v)
        n = sum(literals)
        expr: HplExpression = HplLiteral.number(n)
        for v in variables:
            expr = HplBinaryOperator.addition(v, expr)
        return _simplify(expr)
    if isinstance(arg, HplRange):
        if is_number_literal(arg.min_value) and is_number_literal(arg.max_value):
            n = 0
            lb = int(arg.min_value.value) + (1 if arg.exclude_min else 0)
            ub = int(arg.max_value.value) + (0 if arg.exclude_max else 1)
            for i in range(lb, ub):
                n += i
            return HplLiteral.number(n)
    return call


def _simplify_function_prod(call: HplFunctionCall) -> HplExpression:
    arg: HplExpression = _simplify(call.arguments[0])
    if isinstance(arg, HplSet):
        variables: List[HplExpression] = []
        literals: List[Union[int, float]] = []
        for v in arg.values:
            if is_number_literal(v):
                assert isinstance(v, HplLiteral)
                literals.append(v.value)
            else:
                variables.append(v)
        n = 1
        for v in literals:
            n *= v
        expr: HplExpression = HplLiteral.number(n)
        if n == 0:
            return expr
        for v in variables:
            expr = HplBinaryOperator.multiplication(v, expr)
        return _simplify(expr)
    if isinstance(arg, HplRange):
        if is_number_literal(arg.min_value) and is_number_literal(arg.max_value):
            n = 1
            lb = int(arg.min_value.value) + (1 if arg.exclude_min else 0)
            ub = int(arg.max_value.value) + (0 if arg.exclude_max else 1)
            for i in range(lb, ub):
                n *= i
            return HplLiteral.number(n)
    return call


def _simplify_function_max(call: HplFunctionCall) -> HplExpression:
    if len(call.arguments) == 1:
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplRange):
            if is_number_literal(arg.min_value) and is_number_literal(arg.max_value):
                lb = int(arg.min_value.value) + (1 if arg.exclude_min else 0)
                ub = int(arg.max_value.value) - (1 if arg.exclude_max else 0)
                for i in range(lb, ub):
                    return HplLiteral.number(max(lb, ub))
                # do not perform simplification for empty range
                # FIXME raise error?
            return call
        if isinstance(arg, HplSet):
            values = arg.values
        else:
            return call
    else:
        values = call.arguments

    variables: List[HplExpression] = []
    literals: List[Union[int, float]] = []
    for v in values:
        if is_number_literal(v):
            assert isinstance(v, HplLiteral)
            literals.append(v.value)
        else:
            variables.append(v)
    if len(literals) < 2:
        return call  # nothing to do
    n = HplLiteral.number(max(*literals))
    if not variables:
        return n
    variables.append(n)
    return call.but(arguments=variables)


def _simplify_function_min(call: HplFunctionCall) -> HplExpression:
    if len(call.arguments) == 1:
        arg: HplExpression = _simplify(call.arguments[0])
        if isinstance(arg, HplRange):
            if is_number_literal(arg.min_value) and is_number_literal(arg.max_value):
                lb = int(arg.min_value.value) + (1 if arg.exclude_min else 0)
                ub = int(arg.max_value.value) - (1 if arg.exclude_max else 0)
                for i in range(lb, ub):
                    return HplLiteral.number(min(lb, ub))
                # do not perform simplification for empty range
                # FIXME raise error?
            return call
        if isinstance(arg, HplSet):
            values = arg.values
        else:
            return call
    else:
        values = call.arguments

    variables: List[HplExpression] = []
    literals: List[Union[int, float]] = []
    for v in values:
        if is_number_literal(v):
            assert isinstance(v, HplLiteral)
            literals.append(v.value)
        else:
            variables.append(v)
    if len(literals) < 2:
        return call  # nothing to do
    n = HplLiteral.number(min(*literals))
    if not variables:
        return n
    variables.append(n)
    return call.but(arguments=variables)


def _obviously_different(a: HplExpression, b: HplExpression) -> bool:
    # assume arguments have been simplified
    if _obvious_negatives(a, b):
        return True
    if isinstance(a, HplBinaryOperator):
        op: BinaryOperatorDefinition = a.operator
        assert not isinstance(a.operand1, HplLiteral)  # due to simplification
        if op.is_plus or op.is_minus:
            if a.operand1 == b and isinstance(a.operand2, HplLiteral):
                assert a.operand2.value != 0  # due to simplification
                return True
        if op.is_times:
            if a.operand1 == b and isinstance(a.operand2, HplLiteral):
                assert a.operand2.value != 0  # due to simplification
                assert a.operand2.value != 1  # due to simplification
                return True
        if op.is_division:
            if a.operand1 == b and isinstance(a.operand2, HplLiteral):
                assert a.operand2.value != 0  # due to simplification
                assert a.operand2.value != 1  # due to simplification
                return True
        if op.is_power:
            if a.operand1 == b and isinstance(a.operand2, HplLiteral):
                assert a.operand2.value != 0  # due to simplification
                assert a.operand2.value != 1  # due to simplification
                return True
    return False


def _obvious_negatives(a: HplExpression, b: HplExpression) -> bool:
    # assume arguments have been simplified
    if isinstance(a, HplUnaryOperator):
        if a.operator.is_not or a.operator.is_minus:
            return a.operand == b
    if isinstance(b, HplUnaryOperator):
        if b.operator.is_not or b.operator.is_minus:
            return b.operand == a
    return False


###############################################################################
# Convenience Logic Tests
###############################################################################


def is_not(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 1 and expr.operator.is_not


def is_and(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_and


def is_or(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_or


def is_implies(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_implies


def is_iff(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_iff


def is_inclusion(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_inclusion


def is_comparison(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_comparison


def is_arithmetic_operator(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 2 and expr.operator.is_arithmetic


def is_number_literal(expr: HplExpression) -> bool:
    return expr.is_value and expr.is_literal and expr.can_be_number


def is_negative_number(expr: HplExpression) -> bool:
    return expr.is_operator and expr.arity == 1 and expr.operator.is_minus


def is_true(expr: HplExpression) -> bool:
    return expr.is_value and expr.is_literal and expr.value is True


def is_false(expr: HplExpression) -> bool:
    return expr.is_value and expr.is_literal and expr.value is False


def is_self_or_field(expr: HplExpression, deep: bool = False) -> bool:
    if deep:
        if isinstance(expr, HplUnaryOperator):
            return is_self_or_field(expr.operand, deep=True)
        if isinstance(expr, HplFunctionCall) and expr.arity == 1:
            return is_self_or_field(expr.arguments[0], deep=True)
    return is_self_reference(expr) or (expr.is_accessor and is_self_reference(expr.base_object()))


def is_call_to(expr: HplExpression, function: str) -> bool:
    return expr.is_function_call and expr.function.name == function


def empty_test(expr: HplExpression) -> HplBinaryOperator:
    a = HplFunctionCall('len', (expr,))
    b = HplLiteral('0', 0)
    return HplBinaryOperator('=', a, b)


def inverse_operator(op: BinaryOperatorDefinition) -> BinaryOperatorDefinition:
    inverse = INVERSE_OPERATORS.get(op)
    if inverse is None:
        raise ValueError(f'operator {op!r} does not have an inverse')
    return inverse


def true() -> HplLiteral:
    return HplLiteral('True', True)


def false() -> HplLiteral:
    return HplLiteral('False', False)



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from pytest import raises

from hpl.ast.predicates import HplPredicateExpression
from hpl.errors import HplSanityError, HplSyntaxError
from hpl.parser import condition_parser

###############################################################################
# Test Code
###############################################################################

parser = condition_parser()


def test_addition_is_not_bool():
    with raises(TypeError):
        parser.parse('a + b + c')


def test_set_is_not_bool():
    with raises(TypeError):
        parser.parse('{1} and [2 to f.g.h]')


def test_number_is_not_bool():
    with raises(TypeError):
        parser.parse('(1 and 2) + (not x)')


def test_forall_var_should_be_name():
    with raises(HplSyntaxError):
        parser.parse('forall 42 in D: phi')


def test_forall_domain_should_be_set():
    with raises(TypeError):
        parser.parse('forall x in 42: phi')


def test_forall_condition_should_be_bool():
    with raises(TypeError):
        parser.parse('forall x in D: 42')


def test_bad_syntax_two_binops():
    with raises(HplSyntaxError):
        parser.parse('a implies iff b')


def test_bool_binop_arg_should_be_bool():
    with raises(TypeError):
        parser.parse('a implies 42')


def test_negation_arg_should_be_bool():
    with raises(TypeError):
        parser.parse('not 42')


def test_number_binop_arg_should_be_number():
    with raises(TypeError):
        parser.parse('not a + b')


def test_minus_arg_should_be_number():
    with raises(TypeError):
        parser.parse('-(a and b)')


def test_relational_op_arg_should_be_number():
    with raises(HplSyntaxError):
        parser.parse('a < b < c')
    with raises(TypeError):
        parser.parse('(a < b) < c')


def test_set_is_not_number():
    with raises(TypeError):
        parser.parse('x = -{1,2,3}')


def test_forall_condition_should_include_var_ref():
    with raises(HplSanityError):
        parser.parse('a implies forall x in xs: b')


def test_all_eq_refs_have_same_type():
    with raises(TypeError):
        parser.parse('(a + 1) > 0 and a')
    with raises(TypeError):
        parser.parse('exists x in xs: (a[@x] implies @x)')


def test_at_least_one_self_ref():
    with raises(HplSanityError):
        p = parser.parse('---42 = -42')
        assert isinstance(p, HplPredicateExpression)
        p.check_some_self_references()


def test_unknown_function():
    with raises(ValueError):
        parser.parse('f(x) > 0')



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from hpl.ast.base import HplAstObject
from hpl.ast.events import HplEvent, HplEventDisjunction, HplSimpleEvent
from hpl.ast.expressions import (
    And,
    BuiltinBinaryOperator,
    BuiltinFunction,
    BuiltinUnaryOperator,
    Exists,
    Forall,
    HplArrayAccess,
    HplBinaryOperator,
    HplDataAccess,
    HplExpression,
    HplFieldAccess,
    HplFunctionCall,
    HplLiteral,
    HplQuantifier,
    HplRange,
    HplSet,
    HplThisMessage,
    HplUnaryOperator,
    HplValue,
    HplVarReference,
    Iff,
    Implies,
    Not,
    Or,
    QuantifierType,
)
from hpl.ast.predicates import (
    HplContradiction,
    HplPredicate,
    HplPredicateExpression,
    HplVacuousTruth,
    predicate_from_expression,
)
from hpl.ast.properties import HplPattern, HplProperty, HplScope, PatternType, ScopeType
from hpl.ast.specs import HplSpecification
from hpl.types import (
    ARRAY_TYPE,
    BOOL_TYPE,
    BOOLEANS,
    FLOAT32,
    FLOAT64,
    INT8,
    INT16,
    INT32,
    INT64,
    MESSAGE_TYPE,
    NUMBER_TYPE,
    SET_TYPE,
    STRING_TYPE,
    STRINGS,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    DataType,
    TypeToken,
)



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from hypothesis import assume, given, settings

from hpl.ast import HplAstObject, HplProperty
from hpl.errors import HplSanityError
from hpl.parser import property_parser

from .strategies import properties

###############################################################################
# Property Examples
###############################################################################

GOOD_PROPERTIES = [
    'globally: some topic {int < 1 and float < 2 and string = "hello"}',
    'globally: no topic within 1s',
    'globally: input causes output',
    'globally: input causes output within 1s',
    'globally: output requires input',
    'globally: output requires input within 100 ms',
    (
        'after ~events/bumper {state = PRESSED}:'
        'some ~cmd_vel {linear.x < 0.0 and angular.z = 0.0}'
    ),
    'after input: no output',
    'globally: some topic {m.int in [0 to 10]!}',
    'globally: some topic {not int in [0 to 10]}',
    'globally: some topic {float_array[0] < float_array[1]}',
    'globally: some topic {forall i in [0 to len(int_array)]!: int_array[@i] > 0}',
    'globally: some topic {exists x in int_array: @x > 0}',
    'globally: some topic {len(twist_array) > 0}',
    'until input: some output',
    'after input as M: some output {x = @M.x}',
    'globally: no /joy_teleop/joy {not buttons[0] in {0, 1}}',
    'globally: no /agrob/agrob_mode {not mode in {0,1,2,3}}',
    (
        'after (p1 or p2 as P2 or p3 {phi} or p4 as P4 {phi})'
        'until (q1 or q2 as Q2 or q3 {phi} or q4 as Q4 {phi}):'
        '(a1 or a2 as A2 or a3 {phi} or a4 as A4 {phi})'
        'causes (b1 or b2 as B2 or b3 {phi} or b4 as B4 {phi})'
        'within 100 ms'
    ),
]

###############################################################################
# Test Code
###############################################################################

parser = property_parser()


def test_valid_handwritten_properties():
    for test_str in GOOD_PROPERTIES:
        ast = parser.parse(test_str)
        assert isinstance(ast, HplAstObject)
        assert ast.is_property
        assert isinstance(ast, HplProperty)


@given(properties())
@settings(max_examples=500)
def test_valid_generated_properties(text: str):
    try:
        ast = parser.parse(text)
    except (HplSanityError, TypeError):
        assume(False)  # discard insane properties

    assert isinstance(ast, HplAstObject)
    assert ast.is_property
    assert isinstance(ast, HplProperty)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Dict, List, Mapping, Optional, Set, Tuple

from attrs import field, frozen
from typeguard import typechecked

from hpl.ast.base import HplAstObject
from hpl.ast.expressions import (
    And,
    BuiltinUnaryOperator,
    DataType,
    HplDataAccess,
    HplExpression,
    HplFieldAccess,
    HplLiteral,
    HplValue,
    Not,
)
from hpl.errors import HplSanityError, invalid_type
from hpl.types import TypeToken

###############################################################################
# Top-level Predicate
###############################################################################


@frozen
class HplPredicate(HplAstObject):
    @property
    def is_predicate(self) -> bool:
        return True

    @property
    def is_vacuous(self) -> bool:
        return False

    @property
    def condition(self) -> HplExpression:
        raise NotImplementedError()

    @property
    def phi(self) -> HplExpression:
        return self.condition

    def is_fully_typed(self) -> bool:
        return self.condition.is_fully_typed()

    def negate(self) -> 'HplPredicate':
        raise NotImplementedError()

    def join(self, other: 'HplPredicate') -> 'HplPredicate':
        raise NotImplementedError()

    def external_references(self) -> Set[str]:
        return self.condition.external_references()

    def contains_reference(self, alias: str) -> bool:
        return self.condition.contains_reference(alias)

    def contains_self_reference(self) -> bool:
        return self.condition.contains_self_reference()

    def replace_var_reference(self, alias: str, expr: HplExpression) -> 'HplPredicate':
        raise NotImplementedError()

    def replace_self_reference(self, expr: HplExpression) -> 'HplPredicate':
        raise NotImplementedError()

    def type_check_references(
        self,
        this_msg: TypeToken,
        variables: Optional[Mapping[str, TypeToken]] = None,
    ):
        raise NotImplementedError()


###############################################################################
# General Predicates
###############################################################################


@typechecked
def _cast_expr_to_bool(expr: HplExpression) -> HplExpression:
    return expr.cast(DataType.BOOL)


@frozen
class HplPredicateExpression(HplPredicate):
    expression: HplExpression = field(converter=_cast_expr_to_bool)

    @expression.validator
    def _check_expression(self, _attribute, expr: HplExpression):
        if not expr.can_be_bool:
            raise TypeError(f'not a boolean expression: {{{expr}}}')
        ref_table = _get_reference_table(expr)
        self._all_refs_same_type(ref_table)
        # self._some_field_refs(ref_table)

    def _all_refs_same_type(self, table: Dict[str, List[HplExpression]]):
        # All references to the same field/variable have the same type.
        for ref_group in table.values():
            # must traverse twice, in case we start with the most generic
            # and go down to the most specific
            final_type = DataType.ANY
            for ref in ref_group:
                final_type = ref.data_type.cast(final_type)
            for ref in reversed(ref_group):
                final_type = ref.data_type.cast(final_type)

    def check_some_self_references(self):
        ref_table = _get_reference_table(self.expression)
        self._some_field_refs(ref_table)

    def _some_field_refs(self, table: Dict[str, List[HplExpression]]):
        # There is at least one reference to a field (own).
        #   [NYI] Stricter: one reference per atomic condition.
        for ref_group in table.values():
            for ref in ref_group:
                if not ref.is_accessor:
                    break
                assert isinstance(ref, HplDataAccess)
                if ref.is_indexed:
                    break
                assert isinstance(ref, HplFieldAccess)
                if not ref.message.is_value:
                    break
                assert isinstance(ref.message, HplValue)
                assert ref.message.is_reference
                if not ref.message.is_this_msg:
                    break
                return  # OK
        raise HplSanityError.predicate_without_self_refs(self)

    @property
    def condition(self) -> HplExpression:
        return self.expression

    def children(self) -> Tuple[HplExpression]:
        return (self.expression,)

    def negate(self) -> HplPredicate:
        if self.expression.is_operator:
            if self.expression.operator == BuiltinUnaryOperator.NOT.value:
                return HplPredicateExpression(self.expression.operand)
        return HplPredicateExpression(Not(self.expression))

    def join(self, other: HplPredicate) -> HplPredicate:
        if other.is_vacuous:
            return self if other.is_true else other
        expr = And(self.expression, other.condition)
        return HplPredicateExpression(expr)

    def replace_var_reference(self, alias: str, expr: HplExpression) -> HplPredicate:
        phi: HplExpression = self.expression.replace_var_reference(alias, expr)
        return self.but(expression=phi)

    def replace_self_reference(self, expr: HplExpression) -> HplPredicate:
        phi: HplExpression = self.expression.replace_self_reference(expr)
        return self.but(expression=phi)

    def type_check_references(
        self,
        this_msg: TypeToken,
        variables: Optional[Mapping[str, TypeToken]] = None,
    ):
        return self.expression.type_check_references(this_msg, variables=variables)

    def __str__(self) -> str:
        return f'{{ {self.expression} }}'


###############################################################################
# Vacuous Predicates
###############################################################################


@frozen
class HplVacuousTruth(HplPredicate):
    @property
    def is_vacuous(self) -> bool:
        return True

    @property
    def is_true(self) -> bool:
        return True

    @property
    def condition(self) -> HplExpression:
        return HplLiteral('True', True)

    def is_fully_typed(self) -> bool:
        return True

    def negate(self) -> HplPredicate:
        return HplContradiction()

    def join(self, other: HplPredicate):
        return other

    def external_references(self) -> Set[str]:
        return set()

    def contains_reference(self, _alias: str) -> bool:
        return False

    def contains_self_reference(self) -> bool:
        return False

    def replace_var_reference(self, _alias: str, _expr: HplExpression) -> HplPredicate:
        return self

    def replace_self_reference(self, _expr: HplExpression) -> HplPredicate:
        return self

    def type_check_references(
        self,
        this_msg: TypeToken,
        variables: Optional[Mapping[str, TypeToken]] = None,
    ):
        pass

    def __str__(self) -> str:
        return '{ True }'


@frozen
class HplContradiction(HplPredicate):
    @property
    def is_vacuous(self) -> bool:
        return True

    @property
    def is_true(self) -> bool:
        return False

    @property
    def condition(self) -> HplExpression:
        return HplLiteral('False', False)

    def is_fully_typed(self) -> bool:
        return True

    def negate(self) -> HplPredicate:
        return HplVacuousTruth()

    def join(self, other: HplPredicate) -> HplPredicate:
        return self

    def external_references(self) -> Set[str]:
        return set()

    def contains_reference(self, _alias: str) -> bool:
        return False

    def contains_self_reference(self) -> bool:
        return False

    def replace_var_reference(self, _alias: str, _expr: HplExpression) -> HplPredicate:
        return self

    def replace_self_reference(self, _expr: HplExpression) -> HplPredicate:
        return self

    def type_check_references(
        self,
        this_msg: TypeToken,
        variables: Optional[Mapping[str, TypeToken]] = None,
    ):
        pass

    def __str__(self) -> str:
        return '{ False }'


###############################################################################
# Helper Functions
###############################################################################


def _get_reference_table(expr: HplExpression) -> Dict[str, List[HplExpression]]:
    ref_table = {}
    for obj in expr.iterate():
        assert isinstance(obj, HplExpression)
        if obj.is_accessor or (obj.is_value and obj.is_variable):
            key = str(obj)
            refs = ref_table.get(key)
            if refs is None:
                refs = []
                ref_table[key] = refs
            refs.append(obj)
    return ref_table


def predicate_from_expression(expr: HplExpression) -> HplPredicate:
    if not expr.can_be_bool:
        raise invalid_type('boolean', expr)
    if expr.is_value and expr.is_literal:
        assert isinstance(expr, HplLiteral)
        assert isinstance(expr.value, bool)
        return HplVacuousTruth() if expr.value else HplContradiction()
    return HplPredicateExpression(expr)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Iterator, Mapping, Optional, Set, Tuple

from enum import Enum, auto

from attrs import field, frozen
from attrs.validators import in_, instance_of

from hpl.ast.base import HplAstObject
from hpl.ast.expressions import HplExpression, HplThisMessage
from hpl.ast.predicates import HplPredicate, HplVacuousTruth
from hpl.errors import HplSanityError
from hpl.types import TypeToken

###############################################################################
# Events and Event Operators
###############################################################################


@frozen
class HplEvent(HplAstObject):
    @property
    def is_event(self) -> bool:
        return True

    @property
    def is_simple_event(self) -> bool:
        return False

    @property
    def is_event_disjunction(self) -> bool:
        return False

    def aliases(self) -> Tuple[str]:
        return ()

    def external_references(self) -> Set[str]:
        return set()

    def contains_reference(self, alias: str) -> bool:
        raise NotImplementedError()

    def contains_self_reference(self) -> bool:
        raise NotImplementedError()

    def replace_var_reference(self, alias: str, expr: HplExpression) -> 'HplEvent':
        raise NotImplementedError()

    def simple_events(self) -> Iterator['HplEvent']:
        raise NotImplementedError()

    def type_check_references(self, msg_types: Mapping[str, TypeToken]) -> None:
        raise NotImplementedError()


class EventType(Enum):
    PUBLISH = auto()


@frozen
class HplSimpleEvent(HplEvent):
    name: str
    predicate: HplPredicate = field(validator=instance_of(HplPredicate))
    event_type: EventType = field(validator=in_(EventType))
    alias: Optional[str] = None
    message_type: Optional[TypeToken] = None

    def __attrs_post_init__(self):
        if self.alias:
            phi = self.predicate.replace_var_reference(self.alias, HplThisMessage())
            object.__setattr__(self, 'predicate', phi)

    @property
    def is_simple_event(self) -> bool:
        return True

    @classmethod
    def publish(
        cls, name: str, predicate: Optional[HplPredicate] = None, alias: Optional[str] = None
    ) -> 'HplSimpleEvent':
        if predicate is None:
            predicate = HplVacuousTruth()
        return cls(name, predicate, EventType.PUBLISH, alias=alias)

    @property
    def is_publish(self):
        return self.event_type is EventType.PUBLISH

    @property
    def phi(self) -> HplPredicate:
        return self.predicate

    def children(self) -> Tuple[HplPredicate]:
        return (self.predicate,)

    def aliases(self) -> Tuple[str]:
        if self.alias is None:
            return ()
        return (self.alias,)

    def external_references(self) -> Set[str]:
        refs = self.predicate.external_references()
        if self.alias:
            refs.discard(self.alias)
        return refs

    def contains_reference(self, alias: str) -> bool:
        return self.predicate.contains_reference(alias)

    def contains_self_reference(self) -> bool:
        it_does: bool = self.predicate.contains_self_reference()
        if it_does:
            return True
        return self.alias and self.predicate.contains_reference(self.alias)

    def replace_var_reference(self, alias: str, expr: HplExpression) -> HplEvent:
        phi = self.predicate.replace_var_reference(alias, expr)
        return self.but(predicate=phi)

    def simple_events(self) -> Iterator[HplEvent]:
        yield self

    def type_check_references(self, msg_types: Mapping[str, TypeToken]) -> None:
        this_msg = msg_types[self.name]
        self.predicate.type_check_references(this_msg, variables=msg_types)

    def __str__(self) -> str:
        alias = (' as ' + self.alias) if self.alias is not None else ''
        assert self.is_publish, f'event_type: {self.event_type}'
        return f'{self.name}{alias} {self.predicate}'


@frozen
class HplEventDisjunction(HplEvent):
    event1: HplEvent = field(validator=instance_of(HplEvent))
    event2: HplEvent = field(validator=instance_of(HplEvent))

    def __attrs_post_init__(self):
        names = set()
        pending = list(self.simple_events())
        while pending:
            event = pending.pop()
            assert event.is_simple_event
            if event.name in names:
                raise HplSanityError.duplicate_event(event.name, self)
            names.add(event.name)

    @property
    def is_event_disjunction(self) -> bool:
        return True

    @property
    def events(self) -> Tuple[HplEvent, HplEvent]:
        return (self.event1, self.event2)

    def children(self) -> Tuple[HplEvent, HplEvent]:
        return (self.event1, self.event2)

    def aliases(self) -> Tuple[str]:
        return self.event1.aliases() + self.event2.aliases()

    def external_references(self) -> Set[str]:
        return self.event1.external_references() | self.event2.external_references()

    def contains_reference(self, alias: str) -> bool:
        return self.event1.contains_reference(alias) or self.event2.contains_reference(alias)

    def contains_self_reference(self) -> bool:
        return self.event1.contains_self_reference() or self.event2.contains_self_reference()

    def replace_var_reference(self, alias: str, expr: HplExpression) -> 'HplEventDisjunction':
        e1 = self.event1.replace_var_reference(alias, expr)
        e2 = self.event2.replace_var_reference(alias, expr)
        if e1 is self.event1 and e2 is self.event2:
            return self
        return self.but(event1=e1, event2=e2)

    def simple_events(self) -> Iterator[HplEvent]:
        for event in self.event1.simple_events():
            yield event
        for event in self.event2.simple_events():
            yield event

    def type_check_references(self, msg_types: Mapping[str, TypeToken]) -> None:
        self.event1.type_check_references(msg_types)
        self.event2.type_check_references(msg_types)

    def __str__(self) -> str:
        return f'({self.event1} or {self.event2})'



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any, Final, Iterable, Mapping, Tuple

from enum import Flag, auto

from attrs import field, frozen
from attrs.validators import ge, in_, instance_of

###############################################################################
# Constants
###############################################################################

INF: Final[float] = float('inf')

###############################################################################
# Type System
###############################################################################

# These work as possible types for an expression.
# An expression of unknown type would have "any" type (i.e., all flags on).
# E.g., (NUMBER | BOOL) means the expression can be either a number or a bool.
# Things like variables start with many possible types, and are refined as the
# tree is built.


class DataType(Flag):
    BOOL = auto()
    NUMBER = auto()
    STRING = auto()
    ARRAY = auto()
    RANGE = auto()
    SET = auto()
    MESSAGE = auto()

    NONE = BOOL & NUMBER
    PRIMITIVE = BOOL | NUMBER | STRING
    ITEM = BOOL | NUMBER | STRING | MESSAGE
    COMPOUND = ARRAY | RANGE | SET
    ANY = BOOL | NUMBER | STRING | ARRAY | RANGE | SET | MESSAGE

    @staticmethod
    def union(types: Iterable['DataType']) -> 'DataType':
        result = DataType.NONE
        for t in types:
            result = result | t
        return result

    @property
    def pretty_name(self) -> str:
        ns = []
        for name, member in type(self).__members__.items():
            if self is member:
                return self.name.capitalize()
            if (self & member) != 0:
                ns.append(name)
        return ' or '.join(ns)

    @property
    def can_be_bool(self) -> bool:
        return bool(self & DataType.BOOL)

    @property
    def can_be_number(self) -> bool:
        return bool(self & DataType.NUMBER)

    @property
    def can_be_string(self) -> bool:
        return bool(self & DataType.STRING)

    @property
    def can_be_array(self) -> bool:
        return bool(self & DataType.ARRAY)

    @property
    def can_be_set(self) -> bool:
        return bool(self & DataType.SET)

    @property
    def can_be_range(self) -> bool:
        return bool(self & DataType.RANGE)

    @property
    def can_be_message(self) -> bool:
        return bool(self & DataType.MESSAGE)

    def can_be(self, t: 'DataType') -> bool:
        return bool(self & t)

    def cast(self, t: 'DataType') -> 'DataType':
        r = self & t
        if not r:
            raise TypeError(f"cannot cast '{self}' to '{t}'")
        return r

    def __str__(self) -> str:
        return self.pretty_name


###############################################################################
# Exported Constants
###############################################################################

BOOL_TYPE: Final[DataType] = DataType.BOOL
NUMBER_TYPE: Final[DataType] = DataType.NUMBER
STRING_TYPE: Final[DataType] = DataType.STRING
ARRAY_TYPE: Final[DataType] = DataType.ARRAY
SET_TYPE: Final[DataType] = DataType.SET
MESSAGE_TYPE: Final[DataType] = DataType.MESSAGE

PRIMITIVE_TYPES: Final[Tuple[DataType]] = (BOOL_TYPE, NUMBER_TYPE, STRING_TYPE)

BASE_TYPES: Final[Tuple[DataType]] = (
    BOOL_TYPE,
    NUMBER_TYPE,
    STRING_TYPE,
    ARRAY_TYPE,
    SET_TYPE,
    MESSAGE_TYPE,
)

###############################################################################
# Type Tokens
###############################################################################


@frozen
class TypeToken:
    name: str
    type: DataType = field(validator=in_(BASE_TYPES))

    @property
    def is_bool(self) -> bool:
        return self.type.can_be_bool

    @property
    def is_number(self) -> bool:
        return self.type.can_be_number

    @property
    def is_string(self) -> bool:
        return self.type.can_be_string

    @property
    def is_message(self) -> bool:
        return self.type.can_be_message

    @property
    def is_array(self) -> bool:
        return self.type.can_be_array

    @property
    def is_range(self) -> bool:
        return self.type.can_be_range

    @property
    def is_set(self) -> bool:
        return self.type.can_be_set

    def __str__(self) -> str:
        return self.name


@frozen
class EnumeratedType(TypeToken):
    values: Tuple[Any] = field(factory=tuple, converter=tuple, validator=instance_of(tuple))

    @values.validator
    def _check_values(self, _attribute, values: Tuple[Any]):
        if self.type is DataType.BOOL:
            expected = bool
        elif self.type is DataType.NUMBER:
            expected = (int, float, complex)
        elif self.type is DataType.STRING:
            expected = str
        else:
            expected = object
        for value in values:
            if not isinstance(value, expected):
                raise TypeError(f'{value!r} is not of type {expected}')

    @classmethod
    def booleans(cls, name: str = 'bool') -> 'EnumeratedType':
        return cls(name, type=DataType.BOOL, values=(False, True))


@frozen
class RangedType(TypeToken):
    min_value: Any = field(default=-INF)
    max_value: Any = field(default=INF)

    @max_value.validator
    def _check_max_value(self, _attribute, value: Any):
        if value < self.min_value:
            raise ValueError(f'max_value={value} < min_value={self.min_value}')

    @classmethod
    def uint8(cls, name: str = 'uint8') -> 'RangedType':
        min_value = 0
        max_value = 255
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def uint16(cls, name: str = 'uint16') -> 'RangedType':
        min_value = 0
        max_value = 65535
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def uint32(cls, name: str = 'uint32') -> 'RangedType':
        min_value = 0
        max_value = 4294967295
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def uint64(cls, name: str = 'uint64') -> 'RangedType':
        min_value = 0
        max_value = 18446744073709551615
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def int8(cls, name: str = 'int8') -> 'RangedType':
        min_value = -128
        max_value = 127
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def int16(cls, name: str = 'int16') -> 'RangedType':
        min_value = -32768
        max_value = 32767
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def int32(cls, name: str = 'int32') -> 'RangedType':
        min_value = -2147483648
        max_value = 2147483647
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def int64(cls, name: str = 'int64') -> 'RangedType':
        min_value = -9223372036854775808
        max_value = 9223372036854775807
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def float32(cls, name: str = 'float32') -> 'RangedType':
        min_value = -3.3999999521443642e38
        max_value = 3.3999999521443642e38
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)

    @classmethod
    def float64(cls, name: str = 'float64') -> 'RangedType':
        min_value = -1.7e308
        max_value = 1.7e308
        return cls(name, type=DataType.NUMBER, min_value=min_value, max_value=max_value)


@frozen
class MessageType(TypeToken):
    type: DataType = field(init=False, default=DataType.MESSAGE)
    fields: Mapping[str, TypeToken] = field(factory=dict)
    constants: Mapping[str, Tuple[TypeToken, Any]] = field(factory=dict)

    def leaf_fields(self) -> Mapping[str, TypeToken]:
        fields = {}
        for name, token in self.fields.items():
            if token.is_message:
                for subname, subtoken in token.leaf_fields():
                    fields[f'{name}.{subname}'] = subtoken
            else:
                fields[name] = token
        return fields

    def contains_name(self, name: str) -> bool:
        return name in self.fields or name in self.constants

    def get_type_of(self, name: str) -> TypeToken:
        t: TypeToken = self.fields.get(name)
        return t if t is not None else self.constants[name][0]


@frozen
class ArrayType(TypeToken):
    type: DataType = field(init=False, default=DataType.ARRAY)
    subtype: TypeToken
    length: int = field(default=-1, validator=ge(-1))

    @property
    def is_fixed_length(self) -> bool:
        return self.length >= 0

    def contains_index(self, index: int) -> bool:
        return self.length < 0 or self.length > index


###############################################################################
# Exported Constants
###############################################################################

BOOLEANS: Final[EnumeratedType] = EnumeratedType.booleans()
UINT8: Final[RangedType] = RangedType.uint8()
UINT16: Final[RangedType] = RangedType.uint16()
UINT32: Final[RangedType] = RangedType.uint32()
UINT64: Final[RangedType] = RangedType.uint64()
INT8: Final[RangedType] = RangedType.int8()
INT16: Final[RangedType] = RangedType.int16()
INT32: Final[RangedType] = RangedType.int32()
INT64: Final[RangedType] = RangedType.int64()
FLOAT32: Final[RangedType] = RangedType.float32()
FLOAT64: Final[RangedType] = RangedType.float64()
STRINGS: Final[TypeToken] = TypeToken('string', type=DataType.STRING)



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

PREDICATE_GRAMMAR = r"""
hpl_predicate: "{" condition "}"

hpl_expression: condition

condition: (condition IF_OPERATOR)? disjunction

disjunction: (disjunction OR_OPERATOR)? conjunction

conjunction: (conjunction AND_OPERATOR)? _logic_expr

_logic_expr: negation
           | quantification
           | atomic_condition

negation.5: NOT_OPERATOR _logic_expr

quantification.3: QUANT_OPERATOR CNAME _KW_IN _atomic_value ":" _logic_expr

atomic_condition: expr (RELATIONAL_OPERATOR expr)?

expr: (expr ADD_OPERATOR)? term

term: (term MULT_OPERATOR)? factor

factor: (factor POWER_OPERATOR)? _exponent

_exponent: _atomic_value
         | negative_number
         | "(" condition ")"

negative_number: MINUS_OPERATOR _exponent

_atomic_value: boolean
             | string
             | number_constant
             | number
             | function_call
             | enum_literal
             | range_literal
             | _reference

number_constant: CONSTANT

enum_literal: "{" _enum_member "}"

_enum_member: (_enum_member ",")? expr

range_literal: _start_range expr _KW_TO expr _end_range

_start_range: L_RANGE_EXC | L_RANGE_INC

_end_range: R_RANGE_EXC | R_RANGE_INC

variable: VAR_REF

function_call: CNAME "(" expr ")"

_base_ref: variable
         | own_field

own_field: CNAME

_reference: _base_ref
          | field_access
          | array_access

field_access: _reference "." CNAME

array_access: _reference "[" _index "]"

_index: expr

channel_name: CHANNEL_NAME

int_literal: INT
string: ESCAPED_STRING
number: NUMBER
signed_number: SIGNED_NUMBER
boolean: TRUE | FALSE

TRUE: "True"
FALSE: "False"

RELATIONAL_OPERATOR: EQ_OPERATOR | NEQ_OPERATOR | COMP_OPERATOR | IN_OPERATOR
EQ_OPERATOR: "="
NEQ_OPERATOR: "!="
COMP_OPERATOR: /<=?/ | />=?/
IN_OPERATOR.2: "in"

NOT_OPERATOR.3: "not"
IF_OPERATOR.3: IMPLIES_OPERATOR | IFF_OPERATOR
IMPLIES_OPERATOR.3: "implies"
IFF_OPERATOR.3: "iff"
OR_OPERATOR.3: "or"
AND_OPERATOR.3: "and"

QUANT_OPERATOR.4: ALL_OPERATOR | SOME_OPERATOR
ALL_OPERATOR.4: "forall"
SOME_OPERATOR.4: "exists"

CONSTANT.5: "PI" | "INF" | "NAN" | "E"
ADD_OPERATOR: "+" | "-"
MULT_OPERATOR: "*" | "/"
POWER_OPERATOR: "**"
MINUS_OPERATOR: "-"

L_RANGE_EXC: "!["
L_RANGE_INC: "["
R_RANGE_EXC: "]!"
R_RANGE_INC: "]"

_KW_TO.4: "to"
_KW_IN.4: "in"
_KW_AS.4: "as"
_KW_OR.4: "or"
_KW_WITHIN.4: "within"
_KW_NO.4: "no"
_KW_SOME.4: "some"
_KW_REQUIRES.4: "requires"
_KW_CAUSES.4: "causes"
_KW_FORBIDS.4: "forbids"
_KW_AFTER.4: "after"
_KW_UNTIL.4: "until"
_KW_GLOBALLY.4: "globally"

CHANNEL_NAME: /[\/~]?[a-zA-Z][0-9a-zA-Z_]*(\/[a-zA-Z][0-9a-zA-Z_]*)*/

VAR_REF: "@" CNAME

TIME_UNIT: "s" | "ms"
FREQ_UNIT: "hz"

%import common.CNAME
%import common.INT
%import common.NUMBER
%import common.SIGNED_NUMBER
%import common.ESCAPED_STRING
%import common.WS
%ignore WS

"""

HPL_GRAMMAR = r"""
hpl_file: _list_of_properties

_list_of_properties: (_list_of_properties)? hpl_property

hpl_property: [metadata] _scope ":" _pattern

metadata: _metadata_items

_metadata_items: (_metadata_items)? "#" _metadata_item

_metadata_item: metadata_id
              | metadata_title
              | metadata_desc

metadata_id: "id" ":" CNAME

metadata_title: "title" ":" ESCAPED_STRING

metadata_desc: "description" ":" ESCAPED_STRING

_scope: global_scope
      | after_until
      | until

global_scope: _KW_GLOBALLY

after_until: _KW_AFTER _activator [_KW_UNTIL _terminator]

until: _KW_UNTIL _terminator

_activator: _any_event

_terminator: _any_event

_pattern: existence
        | absence
        | response
        | prevention
        | requirement

existence: _KW_SOME _any_event _time_bound

absence: _KW_NO _any_event _time_bound

response: _any_event _KW_CAUSES _any_event _time_bound

prevention: _any_event _KW_FORBIDS _any_event _time_bound

requirement: _any_event _KW_REQUIRES _any_event _time_bound

_time_bound: [_KW_WITHIN time_amount]

_any_event: event
          | event_disjunction

event: channel_name [alias] [hpl_predicate]

event_disjunction: "(" (event _KW_OR)+ event ")"

time_amount: NUMBER TIME_UNIT

frequency: NUMBER FREQ_UNIT

alias: _KW_AS CNAME

hpl_predicate: "{" condition "}"

hpl_expression: condition

condition: (condition IF_OPERATOR)? disjunction

disjunction: (disjunction OR_OPERATOR)? conjunction

conjunction: (conjunction AND_OPERATOR)? _logic_expr

_logic_expr: negation
           | quantification
           | atomic_condition

negation.5: NOT_OPERATOR _logic_expr

quantification.3: QUANT_OPERATOR CNAME _KW_IN _atomic_value ":" _logic_expr

atomic_condition: expr (RELATIONAL_OPERATOR expr)?

expr: (expr ADD_OPERATOR)? term

term: (term MULT_OPERATOR)? factor

factor: (factor POWER_OPERATOR)? _exponent

_exponent: _atomic_value
         | negative_number
         | "(" condition ")"

negative_number: MINUS_OPERATOR _exponent

_atomic_value: boolean
             | string
             | number_constant
             | number
             | function_call
             | enum_literal
             | range_literal
             | _reference

number_constant: CONSTANT

enum_literal: "{" _enum_member "}"

_enum_member: (_enum_member ",")? expr

range_literal: _start_range expr _KW_TO expr _end_range

_start_range: L_RANGE_EXC | L_RANGE_INC

_end_range: R_RANGE_EXC | R_RANGE_INC

variable: VAR_REF

function_call: CNAME "(" expr ")"

_base_ref: variable
         | own_field

own_field: CNAME

_reference: _base_ref
          | field_access
          | array_access

field_access: _reference "." CNAME

array_access: _reference "[" _index "]"

_index: expr

channel_name: CHANNEL_NAME

int_literal: INT
string: ESCAPED_STRING
number: NUMBER
signed_number: SIGNED_NUMBER
boolean: TRUE | FALSE

TRUE: "True"
FALSE: "False"

RELATIONAL_OPERATOR: EQ_OPERATOR | NEQ_OPERATOR | COMP_OPERATOR | IN_OPERATOR
EQ_OPERATOR: "="
NEQ_OPERATOR: "!="
COMP_OPERATOR: /<=?/ | />=?/
IN_OPERATOR.2: "in"

NOT_OPERATOR.3: "not"
IF_OPERATOR.3: IMPLIES_OPERATOR | IFF_OPERATOR
IMPLIES_OPERATOR.3: "implies"
IFF_OPERATOR.3: "iff"
OR_OPERATOR.3: "or"
AND_OPERATOR.3: "and"

QUANT_OPERATOR.4: ALL_OPERATOR | SOME_OPERATOR
ALL_OPERATOR.4: "forall"
SOME_OPERATOR.4: "exists"

CONSTANT.5: "PI" | "INF" | "NAN" | "E"
ADD_OPERATOR: "+" | "-"
MULT_OPERATOR: "*" | "/"
POWER_OPERATOR: "**"
MINUS_OPERATOR: "-"

L_RANGE_EXC: "!["
L_RANGE_INC: "["
R_RANGE_EXC: "]!"
R_RANGE_INC: "]"

_KW_TO.4: "to"
_KW_IN.4: "in"
_KW_AS.4: "as"
_KW_OR.4: "or"
_KW_WITHIN.4: "within"
_KW_NO.4: "no"
_KW_SOME.4: "some"
_KW_REQUIRES.4: "requires"
_KW_CAUSES.4: "causes"
_KW_FORBIDS.4: "forbids"
_KW_AFTER.4: "after"
_KW_UNTIL.4: "until"
_KW_GLOBALLY.4: "globally"

CHANNEL_NAME: /[\/~]?[a-zA-Z][0-9a-zA-Z_]*(\/[a-zA-Z][0-9a-zA-Z_]*)*/

VAR_REF: "@" CNAME

TIME_UNIT: "s" | "ms"
FREQ_UNIT: "hz"

%import common.CNAME
%import common.INT
%import common.NUMBER
%import common.SIGNED_NUMBER
%import common.ESCAPED_STRING
%import common.WS
%ignore WS

"""

IN_OPERATOR = 'in'
NOT_OPERATOR = 'not'
IMPLIES_OPERATOR = 'implies'
IFF_OPERATOR = 'iff'
OR_OPERATOR = 'or'
AND_OPERATOR = 'and'
ALL_OPERATOR = 'forall'
SOME_OPERATOR = 'exists'



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Tuple

from attrs import field, frozen
from attrs.validators import instance_of

from hpl.ast.base import HplAstObject
from hpl.ast.properties import HplProperty

###############################################################################
# Specifications
###############################################################################


@frozen
class HplSpecification(HplAstObject):
    properties: Tuple[HplProperty] = field(validator=instance_of(tuple))

    @property
    def is_specification(self) -> bool:
        return True

    def children(self) -> Tuple[HplProperty]:
        return self.properties

    def sanity_check(self):
        for prop in self.properties:
            prop.sanity_check()

    def __str__(self) -> str:
        return '\n'.join(str(prop) for prop in self.properties)



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from hypothesis import assume, given, settings

from hpl.ast import HplAstObject, HplExpression
from hpl.parser import expression_parser

from .strategies import expressions

###############################################################################
# Test Code
###############################################################################

parser = expression_parser()


@given(expressions())
@settings(max_examples=500)
# @settings(suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow])
def test_valid_expressions(text: str):
    try:
        ast = parser.parse(text)
        assert isinstance(ast, HplAstObject)
        assert ast.is_expression
        assert isinstance(ast, HplExpression)
    except TypeError:
        assume(False)



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Final, Iterator, Mapping, Optional, Tuple

from enum import Enum, auto

from attrs import field, frozen
from attrs.validators import ge, in_, instance_of, optional

from hpl.ast.base import HplAstObject
from hpl.ast.events import HplEvent, HplSimpleEvent
from hpl.errors import HplSanityError, expected_not_none, invalid_attr
from hpl.types import TypeToken

###############################################################################
# Constants
###############################################################################

INF: Final[float] = float('inf')

###############################################################################
# Scopes
###############################################################################


class ScopeType(Enum):
    GLOBAL = auto()
    AFTER_UNTIL = auto()
    AFTER = auto()
    UNTIL = auto()

    @property
    def is_after(self) -> bool:
        return self is ScopeType.AFTER or self is ScopeType.AFTER_UNTIL

    @property
    def is_until(self) -> bool:
        return self is ScopeType.UNTIL or self is ScopeType.AFTER_UNTIL

    @property
    def is_global(self) -> bool:
        return self is ScopeType.GLOBAL

    @property
    def should_have_activator(self) -> bool:
        return self.is_after

    @property
    def should_have_terminator(self) -> bool:
        return self.is_until


@frozen
class HplScope(HplAstObject):
    scope_type: ScopeType = field(validator=in_(ScopeType))
    activator: Optional[HplEvent] = field(
        default=None,
        validator=optional(instance_of(HplEvent)),
    )
    terminator: Optional[HplEvent] = field(
        default=None,
        validator=optional(instance_of(HplEvent)),
    )

    @activator.validator
    def _check_activator(self, attribute, event: Optional[HplEvent]):
        if self.scope_type.should_have_activator:
            if event is None:
                raise expected_not_none(attribute.name, self)
        else:
            if event is not None:
                raise invalid_attr(attribute.name, None, event, self)

    @terminator.validator
    def _check_terminator(self, attribute, event: Optional[HplEvent]):
        if self.scope_type.should_have_terminator:
            if event is None:
                raise expected_not_none(attribute.name, self)
        else:
            if event is not None:
                raise invalid_attr(attribute.name, None, event, self)

    @property
    def is_scope(self) -> bool:
        return True

    @classmethod
    def globally(cls) -> 'HplScope':
        return cls(ScopeType.GLOBAL)

    @classmethod
    def after(cls, activator: HplEvent) -> 'HplScope':
        return cls(ScopeType.AFTER, activator=activator)

    @classmethod
    def until(cls, terminator: HplEvent) -> 'HplScope':
        return cls(ScopeType.UNTIL, terminator=terminator)

    @classmethod
    def after_until(cls, activator: HplEvent, terminator: HplEvent) -> 'HplScope':
        return cls(ScopeType.AFTER_UNTIL, activator=activator, terminator=terminator)

    @property
    def is_global(self) -> bool:
        return self.scope_type.is_global

    @property
    def is_after(self) -> bool:
        return self.scope_type.is_after

    @property
    def is_until(self) -> bool:
        return self.scope_type.is_until

    @property
    def has_activator(self) -> bool:
        return self.activator is not None

    @property
    def has_terminator(self) -> bool:
        return self.terminator is not None

    def children(self) -> Tuple[HplEvent]:
        if self.activator is None and self.terminator is None:
            return ()
        if self.activator is None:
            return (self.terminator,)
        if self.terminator is None:
            return (self.activator,)
        return (self.activator, self.terminator)

    def __str__(self) -> str:
        if self.scope_type == ScopeType.GLOBAL:
            return 'globally'
        if self.scope_type == ScopeType.AFTER:
            return f'after {self.activator}'
        if self.scope_type == ScopeType.UNTIL:
            return f'until {self.terminator}'
        if self.scope_type == ScopeType.AFTER_UNTIL:
            return f'after {self.activator} until {self.terminator}'
        return self.scope_type.name


###############################################################################
# Patterns
###############################################################################


class PatternType(Enum):
    ABSENCE = auto()
    EXISTENCE = auto()
    REQUIREMENT = auto()
    RESPONSE = auto()
    PREVENTION = auto()

    @property
    def is_safety(self) -> bool:
        return (
            self is PatternType.ABSENCE
            or self is PatternType.REQUIREMENT
            or self is PatternType.PREVENTION
        )

    @property
    def is_liveness(self) -> bool:
        return self is PatternType.EXISTENCE or self is PatternType.RESPONSE

    @property
    def is_absence(self) -> bool:
        return self is PatternType.ABSENCE

    @property
    def is_existence(self) -> bool:
        return self is PatternType.EXISTENCE

    @property
    def is_requirement(self) -> bool:
        return self is PatternType.REQUIREMENT

    @property
    def is_response(self) -> bool:
        return self is PatternType.RESPONSE

    @property
    def is_prevention(self) -> bool:
        return self is PatternType.PREVENTION

    @property
    def should_have_trigger(self) -> bool:
        return (
            self is PatternType.REQUIREMENT
            or self is PatternType.RESPONSE
            or self is PatternType.PREVENTION
        )


@frozen
class HplPattern(HplAstObject):
    pattern_type: PatternType = field(validator=in_(PatternType))
    behaviour: HplEvent = field(validator=instance_of(HplEvent))
    trigger: Optional[HplEvent] = field(default=None, validator=optional(instance_of(HplEvent)))
    min_time: float = field(default=0.0, validator=ge(0.0))
    max_time: float = field(default=INF, converter=float)

    @trigger.validator
    def _check_trigger(self, attribute, event: Optional[HplEvent]):
        if self.pattern_type.should_have_trigger:
            if event is None:
                raise expected_not_none(attribute.name, self)
        else:
            if event is not None:
                raise invalid_attr(attribute.name, None, event, self)

    @max_time.validator
    def _check_max_time(self, attribute, value: float):
        if value < self.min_time:
            raise ValueError(f'{attribute.name}={value!r} < {self.min_time}')

    @property
    def is_pattern(self) -> bool:
        return True

    @classmethod
    def existence(
        cls,
        behaviour: HplEvent,
        min_time: float = 0.0,
        max_time: float = INF,
    ) -> 'HplPattern':
        return cls(
            PatternType.EXISTENCE,
            behaviour,
            trigger=None,
            min_time=min_time,
            max_time=max_time,
        )

    @classmethod
    def absence(
        cls,
        behaviour: HplEvent,
        min_time: float = 0.0,
        max_time: float = INF,
    ) -> 'HplPattern':
        return cls(
            PatternType.ABSENCE,
            behaviour,
            trigger=None,
            min_time=min_time,
            max_time=max_time,
        )

    @classmethod
    def response(
        cls, trigger: HplEvent, response: HplEvent, min_time: float = 0.0, max_time: float = INF
    ) -> 'HplPattern':
        return cls(
            PatternType.RESPONSE,
            response,
            trigger=trigger,
            min_time=min_time,
            max_time=max_time,
        )

    @classmethod
    def requirement(
        cls,
        behaviour: HplEvent,
        requirement: HplEvent,
        min_time: float = 0.0,
        max_time: float = INF,
    ) -> 'HplPattern':
        return cls(
            PatternType.REQUIREMENT,
            behaviour,
            requirement,
            min_time=min_time,
            max_time=max_time,
        )

    @classmethod
    def prevention(
        cls,
        trigger: HplEvent,
        forbidden: HplEvent,
        min_time: float = 0.0,
        max_time: float = INF,
    ) -> 'HplPattern':
        return cls(
            PatternType.PREVENTION,
            forbidden,
            trigger,
            min_time=min_time,
            max_time=max_time,
        )

    @property
    def is_safety(self) -> bool:
        return self.pattern_type.is_safety

    @property
    def is_liveness(self) -> bool:
        return self.pattern_type.is_liveness

    @property
    def is_absence(self) -> bool:
        return self.pattern_type.is_absence

    @property
    def is_existence(self) -> bool:
        return self.pattern_type.is_existence

    @property
    def is_requirement(self) -> bool:
        return self.pattern_type.is_requirement

    @property
    def is_response(self) -> bool:
        return self.pattern_type.is_response

    @property
    def is_prevention(self) -> bool:
        return self.pattern_type.is_prevention

    @property
    def has_min_time(self) -> bool:
        return self.min_time > 0.0 and self.min_time < INF

    @property
    def has_max_time(self) -> bool:
        return self.max_time >= 0.0 and self.max_time < INF

    def children(self) -> Tuple[HplEvent]:
        if self.trigger is None:
            return (self.behaviour,)
        return (self.trigger, self.behaviour)

    def __str__(self) -> str:
        t = ''
        if self.max_time < INF:
            if self.max_time < 1.0:
                t = f' within {self.max_time * 1000}ms'
            else:
                t = f' within {self.max_time}s'
        if self.pattern_type.is_existence:
            return f'some {self.behaviour}{t}'
        if self.pattern_type.is_absence:
            return f'no {self.behaviour}{t}'
        if self.pattern_type.is_response:
            return f'{self.trigger} causes {self.behaviour}{t}'
        if self.pattern_type.is_requirement:
            return f'{self.behaviour} requires {self.trigger}{t}'
        if self.pattern_type.is_prevention:
            return f'{self.trigger} forbids {self.behaviour}{t}'
        return self.pattern_type.name


###############################################################################
# Properties
###############################################################################


@frozen
class HplProperty(HplAstObject):
    scope: HplScope = field(validator=instance_of(HplScope))
    pattern: HplPattern = field(validator=instance_of(HplPattern))

    def __attrs_post_init__(self):
        self.sanity_check()

    @property
    def is_property(self) -> bool:
        return True

    @property
    def is_safety(self) -> bool:
        return self.pattern.is_safety

    @property
    def is_liveness(self) -> bool:
        return self.pattern.is_liveness

    @property
    def uid(self) -> Optional[str]:
        return self.metadata.get('id', None)

    def children(self) -> Tuple[HplScope, HplPattern]:
        return (self.scope, self.pattern)

    def is_fully_typed(self) -> bool:
        for event in self.events():
            for simple_event in event.simple_events():
                e: HplSimpleEvent = simple_event
                if not e.predicate.is_fully_typed():
                    return False
        return True

    def type_check_references(self, msg_types: Mapping[str, TypeToken]) -> None:
        for event in self.events():
            event.type_check_references(msg_types)

    def events(self) -> Iterator[HplEvent]:
        if self.scope.activator is not None:
            yield self.scope.activator
        yield self.pattern.behaviour
        if self.pattern.trigger is not None:
            yield self.pattern.trigger
        if self.scope.terminator is not None:
            yield self.scope.terminator

    def sanity_check(self) -> None:
        initial: Tuple[str] = self._check_activator()
        if self.pattern.is_absence or self.pattern.is_existence:
            self._check_behaviour(initial)
        elif self.pattern.is_requirement:
            aliases = self._check_behaviour(initial)
            self._check_trigger(aliases)
        elif self.pattern.is_response or self.pattern.is_prevention:
            aliases = self._check_trigger(initial)
            self._check_behaviour(aliases)
        else:
            raise TypeError(f'unexpected pattern type: {self.pattern!r}')
        self._check_terminator(initial)

    def _check_activator(self) -> Tuple[str]:
        p = self.scope.activator
        if p is not None:
            self._check_refs_defined(p, ())
            return p.aliases()
        return ()

    def _check_trigger(self, available: Tuple[str]) -> Tuple[str]:
        a = self.pattern.trigger
        assert a is not None
        self._check_refs_defined(a, available)
        aliases = a.aliases()
        self._check_duplicates(aliases, available)
        return aliases + available

    def _check_behaviour(self, available: Tuple[str]) -> Tuple[str]:
        b = self.pattern.behaviour
        self._check_refs_defined(b, available)
        aliases = b.aliases()
        self._check_duplicates(aliases, available)
        return aliases + available

    def _check_terminator(self, available: Tuple[str]) -> None:
        q = self.scope.terminator
        if q is not None:
            self._check_refs_defined(q, available)
            self._check_duplicates(q.aliases(), available)

    def _check_refs_defined(self, event: HplEvent, available: Tuple[str]) -> None:
        for ref in event.external_references():
            if ref not in available:
                raise HplSanityError.ref_undefined_event(ref, event)

    def _check_duplicates(self, aliases: Tuple[str], available: Tuple[str]) -> None:
        for alias in aliases:
            if alias in available:
                raise HplSanityError.already_defined(alias, self)

    def __str__(self) -> str:
        return f'{self.scope}: {self.pattern}'



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

"""
Module that contains the command line program.

Why does this file exist, and why not put this in __main__?

  In some cases, it is possible to import `__main__.py` twice.
  This approach avoids that. Also see:
  https://click.palletsprojects.com/en/5.x/setuptools/#setuptools-integration

Some of the structure of this file came from this StackExchange question:
  https://softwareengineering.stackexchange.com/q/418600
"""

###############################################################################
# Imports
###############################################################################

from typing import Any, Dict, Final, List, Optional

import argparse
from enum import Enum
import json
from math import isinf, isnan
from pathlib import Path
import sys
from traceback import print_exc

from attrs import asdict

from hpl import __version__ as current_version
from hpl.ast.base import HplAstObject
from hpl.errors import HplSyntaxError
from hpl.parser import parse_property, parse_specification

###############################################################################
# Constants
###############################################################################

PROG: Final[str] = 'hpl'

FORMAT_JSON: Final[str] = 'json'

###############################################################################
# Argument Parsing
###############################################################################


def parse_arguments(argv: Optional[List[str]]) -> Dict[str, Any]:
    description = 'Command-line parser for HPL properties.'
    parser = argparse.ArgumentParser(prog=PROG, description=description)

    parser.add_argument(
        '--version',
        action='version',
        version=f'{PROG} {current_version}',
        help='prints the program version',
    )

    parser.add_argument(
        '-p',
        '--property',
        dest='as_property',
        action='store_true',
        help='process arg as HPL property (default: HPL file)',
    )

    parser.add_argument(
        '-o',
        '--output',
        choices=(FORMAT_JSON,),
        help='output format (default: none)',
    )

    parser.add_argument(
        'arg',
        help=f'argument for the {PROG} command',
    )

    args = parser.parse_args(args=argv)
    return vars(args)


###############################################################################
# Setup
###############################################################################


def load_configs(args: Dict[str, Any]) -> Dict[str, Any]:
    try:
        config: Dict[str, Any] = {}
        # with open(args['config_path'], 'r') as file_pointer:
        # yaml.safe_load(file_pointer)

        # arrange and check configs here

        return config
    except Exception as err:
        # log or raise errors
        print(err, file=sys.stderr)
        if str(err) == 'Really Bad':
            raise err

        # Optional: return some sane fallback defaults.
        sane_defaults: Dict[str, Any] = {}
        return sane_defaults


###############################################################################
# Helper Functions
###############################################################################


def _ast_object_serializer(_ast: HplAstObject, _field: Any, value: Any) -> Any:
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, float) and (isinf(value) or isnan(value)):
        return None
    return value


###############################################################################
# Entry Point
###############################################################################


def main(argv: Optional[List[str]] = None) -> int:
    args = parse_arguments(argv)

    try:
        # Load additional config files here, e.g., from a path given via args.
        # Alternatively, set sane defaults if configuration is missing.
        _config = load_configs(args)

        if args['as_property']:
            result: HplAstObject = parse_property(args['arg'])
        else:
            path: Path = Path(args['arg']).resolve(strict=True)
            text: str = path.read_text(encoding='utf-8')
            result = parse_specification(text)

        format: Optional[str] = args.get('output')
        if format == FORMAT_JSON:
            data: Dict[str, Any] = asdict(result, value_serializer=_ast_object_serializer)
            output: str = json.dumps(data, indent=2)
            print(output)

    except HplSyntaxError as hse:
        print('Syntax error:', file=sys.stderr)
        print(hse, file=sys.stderr)
        return 1

    except KeyboardInterrupt:
        print('Aborted manually.', file=sys.stderr)
        return 1

    except Exception as err:
        print('An unhandled exception crashed the application!')
        print(err)
        print_exc()
        return 1

    return 0  # success



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from hpl.ast.expressions import HplBinaryOperator, HplExpression, HplFieldAccess, HplLiteral
from hpl.ast.predicates import HplContradiction, HplPredicate, HplPredicateExpression, HplVacuousTruth
from hpl.parser import condition_parser, expression_parser
from hpl.rewrite import (
    is_true,
    refactor_reference,
    replace_this_with_var,
    replace_var_with_this,
    simplify,
    split_and,
)

###############################################################################
# Predicate Examples
###############################################################################

# Tests will try to move predicate from @A to @B.

NO_REFS = [
    'a + b < c',
    'forall x in xs: @x',
    'not a + b < c',
    '(not ((a or b) implies c) and d)',
    'f.int in [0 to 10]',
    'not array[0] in {0, 1, 2}',
    'x < @Y.y',
    'x = y',
]

REF_BUT_NO_SPLITS = [
    'forall x in @B.xs: xs[@x]',  # reference in domain
    'exists x in xs: @B.x < @x',  # existential quantifier
    'forall x in xs: (@x or @B.x)',  # disjunction
    'forall x in xs: (@x implies @B.x)',  # implication
    'forall x in xs: (@x iff @B.x)',  # equivalence
    'forall x in xs: (not @B.x = @x)',  # negation
    'forall x in xs: (@B.x < @x and @B.y > @x)',  # ref on both sides
    'not forall x in xs: (@x and @B.x)',  # universal quantifier
    'not exists x in xs: (@x and @B.x)',  # negated conjunction
    'not (x < @B.x implies y > @B.y)',  # ref on both sides
    'not (x < @B.x or y > @B.y)',  # ref on both sides
    'not (x and @B.x)',  # negated conjunction
    'x or @B.x',  # disjunction
    'x implies @B.x',  # implication
    'x iff @B.x',  # equivalence
    'x < @B.x and y > @B.y',  # ref on both sides
]

REF_WITH_SPLITS = [
    'x and y > @B.y',  # ref on one side
    'x < @B.x and y',  # ref on one side
    'not (not (x and y > @B.y))',  # double negation
    'not (x or @B.x)',  # negated disjunction
    'not (x implies @B.x)',  # negated implication
    'not exists x in xs: (@x or @B.x)',  # negated disjunction
    'forall x in xs: (@x and @B.x)',  # ref on one side
    'forall x in xs: not (@x or @B.x)',  # ref on one side
    'forall x in xs: (@x >= 0 and @B.x < @x)',  # ref on one side
]

###############################################################################
# Test Code
###############################################################################

parser = condition_parser()


def test_refactor_but_no_references():
    for test_str in NO_REFS:
        predicate: HplPredicate = parser.parse(test_str)
        phi, psi = refactor_reference(predicate.condition, 'B')
        assert phi.is_expression
        assert psi.is_expression
        assert phi is predicate.condition
        assert is_true(psi)


def test_refactor_references_but_no_splits():
    for test_str in REF_BUT_NO_SPLITS:
        predicate: HplPredicate = parser.parse(test_str)
        phi, psi = refactor_reference(predicate.condition, 'B')
        assert phi.is_expression
        assert psi.is_expression
        # assert psi is predicate.condition
        assert is_true(phi)


def test_refactor_references_with_splits():
    for test_str in REF_WITH_SPLITS:
        predicate: HplPredicate = parser.parse(test_str)
        phi, psi = refactor_reference(predicate.condition, 'B')
        assert phi.is_expression
        assert psi.is_expression
        assert phi != predicate.condition
        assert psi != predicate.condition
        assert not is_true(phi)
        assert not is_true(psi)


def test_replace_this_msg_but_no_splits():
    for test_str in REF_BUT_NO_SPLITS:
        predicate: HplPredicate = parser.parse(test_str)
        phi, psi = refactor_reference(predicate.condition, 'B')
        assert is_true(phi)
        replace_this_with_var(psi, 'A')
        replace_var_with_this(psi, 'B')


def test_replace_this_msg_with_splits():
    for test_str in REF_WITH_SPLITS:
        predicate: HplPredicate = parser.parse(test_str)
        phi, psi = refactor_reference(predicate.condition, 'B')
        assert isinstance(phi, HplExpression)
        assert isinstance(psi, HplExpression)
        assert not is_true(phi)
        assert not is_true(psi)
        replace_this_with_var(psi, 'A')
        replace_var_with_this(psi, 'B')


def test_split_conjunction_but_no_splits():
    examples = [
        'a + b < c',  # atomic
        'forall x in xs: @x',  # atomic
        'not a + b < c',  # atomic
        'f.int in [0 to 10]',  # atomic
        'not array[0] in {0, 1, 2}',  # atomic
        'exists x in xs: @B.x < @x',  # existential quantifier
        'forall x in xs: (@x or @B.x)',  # disjunction
        'forall x in xs: (@x implies @B.x)',  # implication
        'forall x in xs: (@x iff @B.x)',  # equivalence
        'forall x in xs: (not @B.x = @x)',  # negation
        'not forall x in xs: (@x and @B.x)',  # negated universal quantifier
        'not exists x in xs: (@x and @B.x)',  # negated conjunction
        'not (x and @B.x)',  # negated conjunction
        'x or @B.x',  # disjunction
        'x implies @B.x',  # implication
        'x iff @B.x',  # equivalence
    ]
    for test_str in examples:
        predicate: HplPredicate = parser.parse(test_str)
        conditions = split_and(predicate.condition)
        assert isinstance(conditions, list)
        assert len(conditions) == 1
        phi = conditions[0]
        assert isinstance(phi, HplExpression)
        # expression must be equivalent, but not exactly equal
        # assert phi == predicate.condition


def test_split_conjunction_with_splits():
    examples = [
        'x < 0 and y > 0',  # conjunction
        'not (x < 0 or y > 0)',  # negated disjunction
        'not (x < 0 implies y > 0)',  # negated implication
        'not exists x in xs: (@x or @B.x)',  # negated existential
        'forall x in xs: (0 < @x and 2 > @x)',  # universal quantifier
        'forall x in xs: not (0 < @x or 2 > @x)',  # universal quantifier
        'forall x in xs: not (0 < @x implies 2 > @x)',  # universal quantifier
        'forall x in xs: forall y in ys: (@x < 0 and @y > 0)',  # nested universal
        'forall x in xs: forall y in ys: not (@x < 0 or @y > 0)',  # nested universal
    ]
    for test_str in examples:
        predicate: HplPredicate = parser.parse(test_str)
        conditions = split_and(predicate.condition)
        assert isinstance(conditions, list)
        assert len(conditions) > 1
        for phi in conditions:
            assert isinstance(phi, HplExpression)
            assert phi != predicate.condition


def test_simplify_vacuous():
    p: HplPredicate = HplVacuousTruth()
    q: HplPredicate = simplify(p)
    assert q is p
    p = HplContradiction()
    q = simplify(p)
    assert q is p
    p = parser.parse('True')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)


def test_simplify_double_negation():
    p: HplPredicate = parser.parse('not not True')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('not not False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('not not a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == parser.parse('a')


def test_simplify_iff():
    p: HplPredicate = parser.parse('True iff True')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('False iff False')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('a iff a')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('True iff False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('False iff True')
    q = simplify(p)
    assert isinstance(q, HplContradiction)


def test_simplify_implies():
    p: HplPredicate = parser.parse('True implies True')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('False implies False')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('True implies False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('False implies True')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('a implies a')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('(not a) implies a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == parser.parse('a')


def test_simplify_and():
    p: HplPredicate = parser.parse('True and True')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('True and False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('a and False')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('False and a')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('(not a) and a')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    p = parser.parse('a and (not a)')
    q = simplify(p)
    assert isinstance(q, HplContradiction)
    a: HplExpression = parser.parse('a')
    p = parser.parse('a and True')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a
    p = parser.parse('True and a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a
    p = parser.parse('a and a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a


def test_simplify_or():
    p: HplPredicate = parser.parse('True or True')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('True or False')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('a or True')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('True or a')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('(not a) or a')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    p = parser.parse('a or (not a)')
    q = simplify(p)
    assert isinstance(q, HplVacuousTruth)
    a: HplExpression = parser.parse('a')
    p = parser.parse('a or False')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a
    p = parser.parse('False or a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a
    p = parser.parse('a or a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == a


def test_simplify_comparison():
    p: HplPredicate = parser.parse('a < 0')
    q: HplPredicate = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == p
    p = parser.parse('0 < a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == parser.parse('a > 0')
    p = parser.parse('0 = a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == parser.parse('a = 0')
    p = parser.parse('a = b')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == p
    p = parser.parse('a + 1 != a')
    q = simplify(p)
    assert isinstance(q, HplPredicateExpression)
    assert q == parser.parse('a != a + 1')


def test_simplify_arithmetic():
    parser = expression_parser()
    p: HplExpression = parser.parse('a = ---1 + 1')
    q: HplExpression = simplify(p)
    assert isinstance(q, HplExpression)
    assert q == parser.parse('a = 0')
    p = parser.parse('0 = ---1 + 1')
    q = simplify(p)
    assert isinstance(q, HplExpression)
    assert q == parser.parse('True')
    p = parser.parse('1 = --1 + 1')
    q = simplify(p)
    assert isinstance(q, HplExpression)
    assert q == parser.parse('False')
    p = parser.parse('((2 - 1) * a ** 1) = b * (1 / b)')
    q = simplify(p)
    assert isinstance(q, HplExpression)
    assert q == parser.parse('a = 1')



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any, Dict, Iterator, Tuple

from attrs import evolve, field, frozen

###############################################################################
# Top-level Classes
###############################################################################


@frozen
class HplAstObject:
    metadata: Dict[str, Any] = field(factory=dict, init=False, eq=False)

    @property
    def is_specification(self) -> bool:
        return False

    @property
    def is_property(self) -> bool:
        return False

    @property
    def is_scope(self) -> bool:
        return False

    @property
    def is_pattern(self) -> bool:
        return False

    @property
    def is_event(self) -> bool:
        return False

    @property
    def is_predicate(self) -> bool:
        return False

    @property
    def is_expression(self) -> bool:
        return False

    def children(self) -> Tuple['HplAstObject']:
        return ()

    def iterate(self) -> Iterator['HplAstObject']:
        stack = [self]
        while stack:
            obj = stack.pop()
            stack.extend(reversed(obj.children()))
            yield obj

    def but(self, **kwargs) -> 'HplAstObject':
        metadata = kwargs.get('metadata')
        if metadata is None:
            for key, value in kwargs.items():
                if getattr(self, key) is not value:
                    break
            else:
                return self  # nothing changes
            metadata = dict(self.metadata)
        new = evolve(self, **kwargs)
        assert new.metadata is not self.metadata
        new.metadata.update(metadata)
        # object.__setattr__(new, 'metadata', metadata)
        return new



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any

###############################################################################
# Functions
###############################################################################


def type_error_in_expr(error: TypeError, expr: Any) -> TypeError:
    return TypeError(f'type error in expression «{expr}»: {error!r}')


def invalid_type(expected: str, found: Any) -> TypeError:
    return TypeError(f'expected type {expected} but found {found!r}')


def missing_field(type_token: Any, field: str, obj: Any) -> TypeError:
    return TypeError(f"type '{type_token}' has no field '{field}' in «{obj!r}»")


def invalid_attr(key: str, expected: Any, found: Any, obj: Any) -> ValueError:
    return ValueError(f'expected {key}={expected!r} but got {found!r} in «{obj!r}»')


def expected_not_none(key: str, obj: Any) -> ValueError:
    return ValueError(f'expected {key} != None in «{obj!r}»')


def index_out_of_range(t_array: Any, idx: int, obj: Any) -> IndexError:
    return IndexError(f"index {idx} out of range in type '{t_array}' in «{obj!r}»")


###############################################################################
# Exceptions
###############################################################################


class HplSanityError(Exception):
    @classmethod
    def predicate_without_self_refs(cls, obj) -> 'HplSanityError':
        return cls(f'there are no references to message fields in «{obj}»')

    @classmethod
    def duplicate_event(cls, name: str, obj) -> 'HplSanityError':
        return cls(f"channel '{name}' appears multiple times in «{obj}»")

    @classmethod
    def ref_undefined_event(cls, name: str, obj) -> 'HplSanityError':
        return cls(f"reference to undefined event '{name}' in «{obj}»")

    @classmethod
    def already_defined(cls, name: str, obj) -> 'HplSanityError':
        return cls(f"multiple definitions of '{name}' in «{obj}»")


class HplSyntaxError(Exception):
    @classmethod
    def duplicate_metadata(cls, key, pid=None):
        which = ''
        if pid is not None:
            which = f" for property '{pid}'"
        return cls(f"duplicate metadata key '{key}'{which}")

    @classmethod
    def from_lark(cls, lark_exception):
        return cls(str(lark_exception))



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from enum import Enum, auto

from attrs import evolve, field, frozen

from hpl.ast.base import HplAstObject
from hpl.ast.expressions import DataType
from hpl.errors import HplTypeError

###############################################################################
# Top-level Classes
###############################################################################


class HplAstObjectType(Enum):
    SPECIFICATION = auto()
    PROPERTY = auto()
    SCOPE = auto()
    PATTERN = auto()
    EVENT = auto()
    PREDICATE = auto()
    EXPRESSION = auto()


@frozen
class HplAstMetadata:
    object_type: HplAstObjectType

    @property
    def is_specification(self) -> bool:
        return self.object_type == HplAstObjectType.SPECIFICATION

    @property
    def is_property(self) -> bool:
        return self.object_type == HplAstObjectType.PROPERTY

    @property
    def is_scope(self) -> bool:
        return self.object_type == HplAstObjectType.SCOPE

    @property
    def is_pattern(self) -> bool:
        return self.object_type == HplAstObjectType.PATTERN

    @property
    def is_event(self) -> bool:
        return self.object_type == HplAstObjectType.EVENT

    @property
    def is_predicate(self) -> bool:
        return self.object_type == HplAstObjectType.PREDICATE

    @property
    def is_expression(self) -> bool:
        return self.object_type == HplAstObjectType.EXPRESSION


###############################################################################
# Expressions
###############################################################################


@frozen
class HplExpression(HplAstObject):
    data_type: DataType = field()

    @data_type.default
    def _get_default_data_type(self):
        return self.default_data_type

    @data_type.validator
    def _check_own_data_type(self, _attribute, value: DataType):
        self.default_data_type.cast(value)

    @property
    def default_data_type(self) -> DataType:
        return DataType.ANY

    def cast(self, t: DataType) -> 'HplExpression':
        try:
            r: DataType = self.data_type.cast(t)
            return evolve(self, data_type=r)
        except TypeError as e:
            raise HplTypeError.in_expr(self, str(e))


###############################################################################
# Predicates
###############################################################################


@frozen
class HplPredicate(HplAstObject):
    condition: HplExpression

    _DIFF_TYPES = "multiple occurrences of '{}' with incompatible types: " "found ({}) and ({})"
    _NO_REFS = "there are no references to any fields of this message"

    def __init__(self, expr):
        if not expr.is_expression:
            raise TypeError("not an expression: " + str(expr))
        if not expr.can_be_bool:
            raise HplTypeError("not a boolean expression: " + str(expr))
        self.condition = expr
        self._static_checks()

    @property
    def phi(self):
        return self.condition

    def is_fully_typed(self):
        return self.condition.is_fully_typed()

    def children(self):
        return (self.condition,)

    def negate(self):
        if self.condition.is_operator and self.condition.operator == "not":
            return HplPredicate(self.condition.operand)
        return HplPredicate(HplUnaryOperator("not", self.condition))

    def join(self, other):
        if other.is_vacuous:
            return self if other.is_true else other
        expr = HplBinaryOperator("and", self.condition, other.condition)
        return HplPredicate(expr)

    def external_references(self):
        return self.condition.external_references()

    def contains_reference(self, alias):
        return self.condition.contains_reference(alias)

    def refine_types(self, rostype, aliases=None):
        # rostype: ROS Type Token
        # aliases: string (alias) -> ROS Type Token
        aliases = aliases if aliases is not None else {}
        stack = [self.condition]
        while stack:
            obj = stack.pop()
            if obj.is_accessor:
                self._refine_type(obj, rostype, aliases)
            else:
                stack.extend(reversed(obj.children()))

    def replace_self_reference(self, alias):
        for obj in self.iterate():
            if obj.is_expression and obj.is_accessor:
                if obj.is_field and obj.message.is_value:
                    if obj.message.is_variable:
                        if obj.message.name == alias:
                            msg = HplThisMessage()
                            obj.message = msg
                            obj._type_check(msg, T_MSG)

    def _refine_type(self, accessor, rostype, aliases):
        stack = [accessor]
        expr = accessor.message
        while expr.is_accessor:
            stack.append(expr)
            expr = expr.message
        assert expr.is_value and (expr.is_this_msg or expr.is_variable)
        if expr.is_this_msg:
            t = rostype
        else:
            if expr.name not in aliases:
                raise HplSanityError("undefined message alias: '{}'".format(expr.name))
            t = aliases[expr.name]
        assert t.is_message
        expr.ros_type = t
        while stack:
            expr = stack.pop()
            if expr.is_field:
                if not (t.is_message or expr.field in t.fields or expr.field in t.constants):
                    raise HplTypeError.ros_field(t, expr.field, expr)
                if expr.field in t.fields:
                    t = t.fields[expr.field]
                else:
                    assert expr.field in t.constants, "'{}' not in {} or {}".format(
                        expr.field, t.fields, t.constants
                    )
                    t = t.constants[expr.field].ros_type
            else:
                assert expr.is_indexed
                if not t.is_array:
                    raise HplTypeError.ros_array(t, expr)
                i = expr.index
                if i.is_value and i.is_literal and not t.contains_index(i.value):
                    raise HplTypeError.ros_index(t, expr.index, expr)
                t = t.type_token
            if t.is_message:
                accessor._type_check(expr, T_MSG)
            elif t.is_array:
                accessor._type_check(expr, T_ARR)
            elif t.is_number:
                accessor._type_check(expr, T_NUM)
                # TODO check that values fit within types
            elif t.is_bool:
                accessor._type_check(expr, T_BOOL)
            elif t.is_string:
                accessor._type_check(expr, T_STR)
            expr.ros_type = t


###############################################################################
# Type Tokens
###############################################################################


@frozen
class TypeToken:
    name: str

    def __attrs_pre_init__(self):
        raise TypeError('cannot instantiate an abstract class')

    @property
    def type(self) -> DataType:
        raise NotImplementedError()

    @property
    def is_bool(self) -> bool:
        return self.type.can_be_bool

    @property
    def is_number(self) -> bool:
        return self.type.can_be_number

    @property
    def is_string(self) -> bool:
        return self.type.can_be_string

    @property
    def is_message(self) -> bool:
        return self.type.can_be_message

    @property
    def is_array(self) -> bool:
        return self.type.can_be_array

    @property
    def is_range(self) -> bool:
        return self.type.can_be_range

    @property
    def is_set(self) -> bool:
        return self.type.can_be_set


@frozen
class StringType(TypeToken):
    @property
    def type(self) -> DataType:
        return DataType.STRING


@frozen
class EnumeratedType(TypeToken):
    _type: DataType = field(validator=in_(PRIMITIVE_TYPES))
    values: Tuple[Any] = field(factory=tuple, converter=tuple, validator=instance_of(tuple))

    @values.validator
    def _check_values(self, _attribute, values: Tuple[Any]):
        if self._type is DataType.BOOL:
            expected = bool
        elif self._type is DataType.NUMBER:
            expected = PY_NUMBER
        elif self._type is DataType.STRING:
            expected = str
        else:
            raise ValueError(f'unexpected type {self._type}')
        for value in values:
            if not isinstance(value, expected):
                raise TypeError(f'{value!r} is not of type {expected}')

    @property
    def type(self) -> DataType:
        return self._type

    @classmethod
    def booleans(cls, name: str = 'bool') -> 'EnumeratedType':
        return cls(name, type=DataType.BOOL, values=(False, True))


@frozen
class NumericType(TypeToken):
    min_value: Union[int, float, complex] = field(default=-INF, validator=instance_of(PY_NUMBER))
    max_value: Union[int, float, complex] = field(default=INF, validator=instance_of(PY_NUMBER))

    @max_value.validator
    def _check_max_value(self, _attribute, value: Union[int, float]):
        if value < self.min_value:
            raise ValueError(f'max_value={value} < min_value={self.min_value}')

    @property
    def type(self) -> DataType:
        return DataType.NUMBER

    @classmethod
    def uint8(cls, name: str = 'uint8') -> 'NumericType':
        min_value = 0
        max_value = 255
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def uint16(cls, name: str = 'uint16') -> 'NumericType':
        min_value = 0
        max_value = 65535
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def uint32(cls, name: str = 'uint32') -> 'NumericType':
        min_value = 0
        max_value = 4294967295
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def uint64(cls, name: str = 'uint64') -> 'NumericType':
        min_value = 0
        max_value = 18446744073709551615
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def int8(cls, name: str = 'int8') -> 'NumericType':
        min_value = -128
        max_value = 127
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def int16(cls, name: str = 'int16') -> 'NumericType':
        min_value = -32768
        max_value = 32767
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def int32(cls, name: str = 'int32') -> 'NumericType':
        min_value = -2147483648
        max_value = 2147483647
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def int64(cls, name: str = 'int64') -> 'NumericType':
        min_value = -9223372036854775808
        max_value = 9223372036854775807
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def float32(cls, name: str = 'float32') -> 'NumericType':
        min_value = -3.3999999521443642e38
        max_value = 3.3999999521443642e38
        return cls(name, min_value=min_value, max_value=max_value)

    @classmethod
    def float64(cls, name: str = 'float64') -> 'NumericType':
        min_value = -1.7e308
        max_value = 1.7e308
        return cls(name, min_value=min_value, max_value=max_value)


@frozen
class MessageType(TypeToken):
    fields: Mapping[str, TypeToken] = field(factory=dict)
    constants: Mapping[str, Any] = field(factory=dict)

    @property
    def type(self) -> DataType:
        return DataType.MESSAGE

    def leaf_fields(self) -> Mapping[str, TypeToken]:
        fields = {}
        for name, token in self.fields.items():
            if token.is_message:
                for subname, subtoken in token.leaf_fields():
                    fields[f'{name}.{subname}'] = subtoken
            else:
                fields[name] = token
        return fields


@frozen
class ArrayType(TypeToken):
    subtype: TypeToken
    length: int = field(default=-1, validator=ge(-1))

    @property
    def type(self) -> DataType:
        return DataType.ARRAY

    @property
    def is_fixed_length(self) -> bool:
        return self.length >= 0


###############################################################################
# Exported Constants
###############################################################################

BOOLEANS: Final[EnumeratedType] = EnumeratedType.booleans()
UINT8: Final[NumericType] = NumericType.uint8()
UINT16: Final[NumericType] = NumericType.uint16()
UINT32: Final[NumericType] = NumericType.uint32()
UINT64: Final[NumericType] = NumericType.uint64()
INT8: Final[NumericType] = NumericType.int8()
INT16: Final[NumericType] = NumericType.int16()
INT32: Final[NumericType] = NumericType.int32()
INT64: Final[NumericType] = NumericType.int64()
FLOAT32: Final[NumericType] = NumericType.float32()
FLOAT64: Final[NumericType] = NumericType.float64()
STRINGS: Final[StringType] = StringType('string')



# SPDX-License-Identifier: MIT
# Copyright © 2021 André Santos

###############################################################################
# Imports
###############################################################################

from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union

from enum import Enum
import math

from attrs import frozen
from lark import Lark, Transformer
from lark.exceptions import UnexpectedCharacters, UnexpectedToken
from lark.visitors import v_args

from hpl.ast import (
    HplArrayAccess,
    HplBinaryOperator,
    HplEventDisjunction,
    HplFieldAccess,
    HplFunctionCall,
    HplLiteral,
    HplPattern,
    HplPredicate,
    HplProperty,
    HplQuantifier,
    HplRange,
    HplScope,
    HplSet,
    HplSimpleEvent,
    HplSpecification,
    HplThisMessage,
    HplUnaryOperator,
    HplVacuousTruth,
    HplVarReference,
)
from hpl.ast.base import HplAstObject
from hpl.ast.events import HplEvent
from hpl.ast.expressions import HplExpression, _convert_binary_operator, _convert_unary_operator
from hpl.ast.predicates import predicate_from_expression
from hpl.errors import HplSyntaxError
from hpl.grammar import HPL_GRAMMAR, PREDICATE_GRAMMAR
from hpl.types import ARRAY_TYPE, MESSAGE_TYPE, NUMBER_TYPE

###############################################################################
# Constants
###############################################################################

INF = float('inf')
NAN = float('nan')


class NumberConstants(Enum):
    PI = math.pi
    INF = INF
    NAN = NAN
    E = math.e


###############################################################################
# Transformer
###############################################################################


@v_args(inline=True)
class PropertyTransformer(Transformer):
    @v_args(inline=False)
    def hpl_file(self, children: Iterable[HplProperty]) -> HplSpecification:
        return HplSpecification(tuple(children))

    def hpl_property(
        self,
        metadata: Optional[Dict[str, Any]],
        scope: HplScope,
        pattern: HplPattern,
    ) -> HplProperty:
        if metadata is None:
            metadata = {}
        # hpl_property.sanity_check()
        hpl_property = HplProperty(scope, pattern)
        hpl_property.metadata.update(metadata)
        return hpl_property

    @v_args(inline=False)
    def metadata(self, children: Iterable[Tuple[str, Any]]) -> Dict[str, Any]:
        metadata: Dict[str, Any] = {}
        pid = None
        dup = None
        for key, value in children:
            if key == 'id':
                pid = value
            if key in metadata:
                dup = key
            metadata[key] = value
        if dup is not None:
            raise HplSyntaxError.duplicate_metadata(dup, pid=pid)
        return metadata

    def metadata_id(self, data: str) -> Tuple[str, str]:
        return ('id', data)

    def metadata_title(self, data: str) -> Tuple[str, str]:
        return ('title', data)

    def metadata_desc(self, data: str) -> Tuple[str, str]:
        return ('description', data)

    @v_args(inline=False)
    def global_scope(self, children: Iterable[Any]) -> HplScope:
        assert not children
        return HplScope.globally()

    def after_until(self, p: HplEvent, q: Optional[HplEvent] = None) -> HplScope:
        return HplScope.after(p) if q is None else HplScope.after_until(p, q)

    def until(self, event: HplEvent) -> HplScope:
        return HplScope.until(event)

    def existence(self, b: HplEvent, t: Optional[float]) -> HplPattern:
        max_time = INF if t is None else t
        return HplPattern.existence(b, max_time=max_time)

    def absence(self, b: HplEvent, t: Optional[float]) -> HplPattern:
        max_time = INF if t is None else t
        return HplPattern.absence(b, max_time=max_time)

    def response(self, a: HplEvent, b: HplEvent, t: Optional[float]) -> HplPattern:
        max_time = INF if t is None else t
        return HplPattern.response(a, b, max_time=max_time)

    def prevention(self, a: HplEvent, b: HplEvent, t: Optional[float]) -> HplPattern:
        max_time = INF if t is None else t
        return HplPattern.prevention(a, b, max_time=max_time)

    def requirement(self, b: HplEvent, a: HplEvent, t: Optional[float]) -> HplPattern:
        max_time = INF if t is None else t
        return HplPattern.requirement(b, a, max_time=max_time)

    @v_args(inline=False)
    def event_disjunction(self, children: Iterable[HplEvent]) -> HplEventDisjunction:
        assert len(children) >= 2
        if len(children) == 2:
            return HplEventDisjunction(children[0], children[1])
        else:
            return HplEventDisjunction(children[0], self.event_disjunction(children[1:]))

    def event(self, name: str, alias: Optional[str], phi: Optional[HplPredicate]) -> HplSimpleEvent:
        phi = HplVacuousTruth() if phi is None else phi
        return HplSimpleEvent.publish(name, alias=alias, predicate=phi)

    def alias(self, name: str) -> str:
        return name

    def hpl_predicate(self, expr: HplExpression) -> HplPredicate:
        return predicate_from_expression(expr)

    def hpl_expression(self, expr: HplExpression) -> HplExpression:
        return expr

    @v_args(inline=False)
    def condition(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    @v_args(inline=False)
    def disjunction(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    @v_args(inline=False)
    def conjunction(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    def negation(self, token: str, phi: HplExpression) -> HplUnaryOperator:
        op = _convert_unary_operator(token)
        return HplUnaryOperator(op, phi.cast(op.parameter))

    def quantification(
        self,
        quantifier: str,
        variable: str,
        domain: HplExpression,
        condition: HplExpression,
    ) -> HplQuantifier:
        return HplQuantifier(quantifier, variable, domain, condition)

    @v_args(inline=False)
    def atomic_condition(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    def function_call(self, fun: str, arg: HplExpression) -> HplExpression:
        return HplFunctionCall(fun, (arg,))

    @v_args(inline=False)
    def expr(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    @v_args(inline=False)
    def term(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    @v_args(inline=False)
    def factor(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        return self._lr_binop(children)

    def _lr_binop(self, children: Iterable[Union[str, HplExpression]]) -> HplExpression:
        assert len(children) == 1 or len(children) == 3
        if len(children) == 3:
            op = _convert_binary_operator(children[1])
            lhs = children[0].cast(op.parameter1)
            rhs = children[2].cast(op.parameter2)
            return HplBinaryOperator(op, lhs, rhs)
        return children[0]  # len(children) == 1

    def negative_number(self, token: str, n: HplExpression) -> HplUnaryOperator:
        op = _convert_unary_operator(token)
        return HplUnaryOperator(op, n.cast(op.parameter))

    def number_constant(self, token: str) -> HplLiteral:
        return HplLiteral(token, NumberConstants[token].value)

    @v_args(inline=False)
    def enum_literal(self, values: Iterable[HplExpression]) -> HplSet:
        return HplSet(values)

    def range_literal(self, lr: str, lb: HplExpression, ub: HplExpression, rr: str) -> HplRange:
        exc_min = lr.startswith('!')
        exc_max = rr.endswith('!')
        return HplRange(lb, ub, exclude_min=exc_min, exclude_max=exc_max)

    def variable(self, token: str) -> HplVarReference:
        return HplVarReference(token)

    def own_field(self, token: str) -> HplFieldAccess:
        return HplFieldAccess(HplThisMessage(), token)

    def field_access(self, ref: HplExpression, token: str) -> HplFieldAccess:
        return HplFieldAccess(ref.cast(MESSAGE_TYPE), token)

    def array_access(self, ref: HplExpression, index: HplExpression) -> HplArrayAccess:
        return HplArrayAccess(ref.cast(ARRAY_TYPE), index.cast(NUMBER_TYPE))

    def frequency(self, num: str, unit: str) -> float:
        n = float(num)
        assert unit == 'hz'
        n = 1.0 / n  # seconds
        return n

    def time_amount(self, num: str, unit: str) -> float:
        n = float(num)
        if unit == 'ms':
            n = n / 1000.0
        else:
            assert unit == 's'
        return n

    def boolean(self, token: str) -> HplLiteral:
        if token == 'True':
            return HplLiteral(token, True)
        assert token == 'False'
        return HplLiteral(token, False)

    def string(self, token: str) -> HplLiteral:
        return HplLiteral(token, token)

    def number(self, token: str) -> HplLiteral:
        try:
            return HplLiteral(token, int(token))
        except ValueError:
            return HplLiteral(token, float(token))

    def signed_number(self, token: str) -> HplLiteral:
        try:
            return HplLiteral(token, int(token))
        except ValueError:
            return HplLiteral(token, float(token))

    def int_literal(self, token: str) -> HplLiteral:
        return HplLiteral(token, int(token))

    def channel_name(self, name: str) -> str:
        return name


###############################################################################
# HPL Parser
###############################################################################


@frozen
class HplParser:
    _lark: Lark
    transform: Optional[Callable[[HplAstObject], HplAstObject]] = None

    @classmethod
    def from_grammar(
        cls,
        grammar: str,
        start: str = 'hpl_file',
        *,
        transform: Optional[Callable[[HplAstObject], HplAstObject]] = None,
        debug: bool = False,
    ) -> 'HplParser':
        return cls(
            Lark(
                grammar,
                parser='lalr',
                start=start,
                transformer=PropertyTransformer(),
                maybe_placeholders=True,
                debug=debug,
            ),
            transform=transform,
        )

    def parse(self, text: str) -> HplAstObject:
        try:
            result: HplAstObject = self._lark.parse(text)
        except (UnexpectedToken, UnexpectedCharacters, SyntaxError) as e:
            raise HplSyntaxError.from_lark(e)
        return result if self.transform is None else self.transform(result)

    @classmethod
    def specification_parser(cls, *, debug: bool = False) -> 'HplParser':
        return cls.from_grammar(HPL_GRAMMAR, start='hpl_file', debug=debug)

    @classmethod
    def property_parser(cls, *, debug: bool = False) -> 'HplParser':
        return cls.from_grammar(HPL_GRAMMAR, start='hpl_property', debug=debug)

    @classmethod
    def predicate_parser(cls, *, debug: bool = False) -> 'HplParser':
        return cls.from_grammar(PREDICATE_GRAMMAR, start='hpl_predicate', debug=debug)

    @classmethod
    def condition_parser(cls, *, debug: bool = False) -> 'HplParser':
        return cls.from_grammar(
            PREDICATE_GRAMMAR,
            start='hpl_expression',
            debug=debug,
            transform=predicate_from_expression,
        )

    @classmethod
    def expression_parser(cls, *, debug: bool = False) -> 'HplParser':
        return cls.from_grammar(PREDICATE_GRAMMAR, start='hpl_expression', debug=debug)


def specification_parser(debug: bool = False) -> HplParser:
    return HplParser.specification_parser(debug=debug)


def property_parser(debug: bool = False) -> HplParser:
    return HplParser.property_parser(debug=debug)


def predicate_parser(debug: bool = False) -> HplParser:
    return HplParser.predicate_parser(debug=debug)


def condition_parser(debug: bool = False) -> HplParser:
    return HplParser.condition_parser(debug=debug)


def expression_parser(debug: bool = False) -> HplParser:
    return HplParser.expression_parser(debug=debug)


def parse_specification(hpl_text: str) -> HplSpecification:
    return specification_parser().parse(hpl_text)


def parse_property(hpl_text: str) -> HplProperty:
    return property_parser().parse(hpl_text)


def parse_predicate(hpl_text: str) -> HplPredicate:
    return predicate_parser().parse(hpl_text)


def parse_condition(hpl_text: str) -> HplPredicate:
    return condition_parser().parse(hpl_text)


def parse_expresion(hpl_text: str) -> HplExpression:
    return expression_parser().parse(hpl_text)



# SPDX-License-Identifier: MIT
# Copyright © 2023 André Santos

###############################################################################
# Imports
###############################################################################

from hpl.ast.properties import HplProperty
from hpl.parser import property_parser
from hpl.rewrite import canonical_form

###############################################################################
# Predicate Examples
###############################################################################


###############################################################################
# Test Code
###############################################################################

parser = property_parser()


def test_canonical_global_absence_no_splits():
    property = parser.parse('globally: no a')
    outputs = canonical_form(property)

    assert len(outputs) == 1
    assert outputs[0] is property


def test_canonical_global_absence_splits():
    property = parser.parse('globally: no (a or b)')
    outputs = canonical_form(property)

    assert len(outputs) == 2

    assert outputs[0].pattern.behaviour.is_simple_event
    assert outputs[1].pattern.behaviour.is_simple_event

    assert outputs[0].pattern.behaviour is property.pattern.behaviour.event1
    assert outputs[1].pattern.behaviour is property.pattern.behaviour.event2


def test_canonical_global_existence_no_splits():
    property = parser.parse('globally: some (a or b)')
    outputs = canonical_form(property)

    assert len(outputs) == 1
    assert outputs[0] is property


def test_canonical_after_existence_splits():
    property = parser.parse('after (a or b): some (c or d)')
    outputs = canonical_form(property)

    assert len(outputs) == 2

    assert outputs[0].scope.activator.is_simple_event
    assert outputs[1].scope.activator.is_simple_event

    assert outputs[0].scope.activator is property.scope.activator.event1
    assert outputs[1].scope.activator is property.scope.activator.event2

    assert outputs[0].pattern is property.pattern
    assert outputs[1].pattern is property.pattern


def test_canonical_global_response_no_splits():
    property = parser.parse('globally: a causes (b or c)')
    outputs = canonical_form(property)

    assert len(outputs) == 1
    assert outputs[0] is property


def test_canonical_global_response_splits():
    property = parser.parse('globally: (a or b) causes (c or d)')
    outputs = canonical_form(property)

    assert len(outputs) == 2

    assert outputs[0].pattern.trigger.is_simple_event
    assert outputs[1].pattern.trigger.is_simple_event

    assert outputs[0].pattern.trigger is property.pattern.trigger.event1
    assert outputs[1].pattern.trigger is property.pattern.trigger.event2
